/*
  Copyright (C) 2002 Renaud Pawlak <renaud@aopsys.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package org.objectweb.jac.aspects.gui;

import org.objectweb.jac.core.Collaboration;
import org.objectweb.jac.core.Display;
import org.objectweb.jac.core.rtti.AbstractMethodItem;
import org.objectweb.jac.core.rtti.ConstructorItem;
import org.objectweb.jac.core.rtti.MethodItem;
import org.objectweb.jac.util.Exceptions;
import org.objectweb.jac.util.Log;
import org.objectweb.jac.util.WrappedThrowableException;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;
import java.util.Iterator;

/**
 * This class allows the programmer to invoke a given method in a new
 * thread.
 *
 * <p>JAC programmers should use JAC to pass some attibutes of the
 * current thread to the new thread.
 *
 * <p>Typical use:
 * <ul><pre>
 * InvokeThread.run( myObject, "myMethodName",
 *                   new Object[] { arg0, ...},
 *                   new String[] { attrName0, ... },
 *                   new Object[] { attrValue0, ... },
 *                   new String[] { lattrName0, ... },
 *                   new Object[] { lattrValue0, ... } );
 * </pre></ul> */

public class InvokeThread extends Thread {
   AbstractMethodItem method;
   Object object;
   Object[] parameters;
   Object returnValue;
   String[] attrNames;
   Object[] attrValues; 
   String[] lattrNames;
   Object[] lattrValues;
   Collaboration parentCollaboration;
   boolean showResult = true;
   //String applicationName;

   /**
    * Runs a method in a new thread and sets a display for this
    * thread.
    *
    * @param object the object that contains the method to call
    * @param method the method to call
    * @param parameters the parameters to pass to the method
    * @param attrNames the attribute names to set into the new thread
    * collaboration
    * @param attrValues the values of these attributes
    * @param lattrNames the local attribute names to set into the new
    * thread collaboration
    * @param lattrValues the values of these local attributes */


   public static InvokeThread run(Object object, AbstractMethodItem method,
                                  Object[] parameters, 
                                  String[] attrNames, Object[] attrValues,
                                  String[] lattrNames, Object[] lattrValues) {
      InvokeThread ret = new InvokeThread( object, method, parameters );
      ret.parentCollaboration = Collaboration.get();
      ret.attrNames = attrNames;
      ret.attrValues = attrValues;
      ret.lattrNames = lattrNames;
      ret.lattrValues = lattrValues;
      ret.start();
      return ret;
   }

   /**
    * Runs a method in a new thread and sets a display for this
    * thread. Do not show any results on the display.
    *
    * @param object the object that contains the method to call
    * @param method the method to call
    * @param parameters the parameters to pass to the method
    * @param attrNames the attribute names to set into the new thread
    * collaboration
    * @param attrValues the values of these attributes
    * @param lattrNames the local attribute names to set into the new
    * thread collaboration
    * @param lattrValues the values of these local attributes */

   public static 
   InvokeThread quietRun(Object object, 
                         AbstractMethodItem method,
                         Object[] parameters, 
                         String[] attrNames, Object[] attrValues,
                         String[] lattrNames, Object[] lattrValues ) 
   {
      InvokeThread ret = new InvokeThread(object,method,parameters);
      ret.parentCollaboration = Collaboration.get();
      ret.attrNames = attrNames;
      ret.attrValues = attrValues;
      ret.lattrNames = lattrNames;
      ret.lattrValues = lattrValues;
      ret.showResult = false;
      ret.start();
      return ret;
   }

   /**
    * Runs a method in a new thread.
    *
    * @param object the object that contains the method to call
    * @param method the method to call
    * @param parameters the parameters to pass to the method */

   public static InvokeThread run(Object object, AbstractMethodItem method,
                                  Object[] parameters) {
      InvokeThread ret = new InvokeThread( object, method, parameters );
      ret.start();
      return ret;
   }

   /**
    * Creates a new thread that will invoke a method when started.
    *
    * <p>The programmer should use the static <code>run</code>
    * methods.
    *
    * @param object the object that contains the method to invoke
    * @param method the method to invoke when started
    * @param parameters the parameters to pass to the method */

   public InvokeThread( Object object, AbstractMethodItem method,
                        Object[] parameters ) {
      this.method = method;
      this.object = object;
      this.parameters = parameters;
   }


   public InvokeThread(Object object, AbstractMethodItem method,
                       Object[] parameters, 
                       String[] attrNames, Object[] attrValues,
                       String[] lattrNames, Object[] lattrValues) {
      this.method = method;
      this.object = object;
      this.parameters = parameters;
      this.parentCollaboration = Collaboration.get();
      this.attrNames = attrNames;
      this.attrValues = attrValues;
      this.lattrNames = lattrNames;
      this.lattrValues = lattrValues;
      this.showResult = false;
   }

   /**
    * Runs the thread (and invoke the method that was given to the
    * constructor with the right display in the collaboration).
    *
    * <p>Do not call this method directly. */

   public void run() {
      Collaboration collab = Collaboration.get();
      Log.trace("threads","invokeThread "+this+": "+
                object+"."+method+Arrays.asList(parameters));
      if (parentCollaboration!=null) {
         Iterator it = parentCollaboration.attributeNames().iterator();
         while (it.hasNext()) {
            String attrName = (String)it.next();
            collab.addAttribute(
               attrName,parentCollaboration.getAttribute(attrName));
         }         
         Log.trace("threads","application = "+parentCollaboration.getCurApp());
         collab.setCurApp(parentCollaboration.getCurApp());
      }
      if (attrNames != null) {
         for (int i=0; i<attrNames.length; i++) {
            Log.trace("threads",
                      "setting attribute " + attrNames[i] +
                      " to " + attrValues[i]);
            collab.addAttribute( 
               attrNames[i], attrValues[i]);
         }
      }
      if (lattrNames != null) {
         for (int i=0; i<lattrNames.length; i++) {
            Log.trace("threads",
                      "setting local attribute " + 
                      lattrNames[i] + " to " + lattrValues[i]);
            collab.addAttribute( 
               lattrNames[i], lattrValues[i]);
         }
      }
      DisplayContext context = (DisplayContext)collab
         .getAttribute(GuiAC.DISPLAY_CONTEXT);

      CustomizedDisplay display = context.getDisplay();
      try {
         Log.trace("threads","InvokeThread " + method + "," + object + "," + 
                   Arrays.asList(parameters));
         
         Class[] paramTypes = method.getParameterTypes();
         if (paramTypes.length != parameters.length) 
            throw new RuntimeException("Wrong number of parameters ("+
                                       parameters.length+") for "+method);
         for (int i=0; i< parameters.length; i++) {
            if (parameters[i]==null) {
               if (paramTypes[i] == float.class) {
                  parameters[i] = new Float(0.0);
               } else if (paramTypes[i] == long.class) {
                  parameters[i] = new Long(0);
               } else if (paramTypes[i] == double.class) {
                  parameters[i] = new Double(0.0);
               } else if (paramTypes[i] == byte.class) {
                  parameters[i] = new Byte((byte)0);
               } else if (paramTypes[i] == char.class) {
                  parameters[i] = new Character(' ');
               } else if (paramTypes[i] == short.class) {
                  parameters[i] = new Short((short)0);
               } else if (paramTypes[i] == int.class) {
                  parameters[i] = new Integer(0);
               } else if (paramTypes[i] == boolean.class) {
                  parameters[i] = Boolean.FALSE;
               }
            }
         }

         if ( method instanceof ConstructorItem ) {
            returnValue = ((ConstructorItem)method).newInstance(parameters);
         } else {
            returnValue = ((MethodItem)method).invoke(object, parameters);
         }

         if (method.getType() != void.class) {
            if (display != null && showResult) {

               if (collab.getAttribute(GuiAC.OPEN_VIEW)!=null) {
                  display.openView(returnValue);
               } else {
		   if(method.getAttribute(GuiAC.SMALL_TARGET_CONTAINER)!=null) {
		       EventHandler.get().onSelection(context,method,returnValue,null,null,false);
		   } else {
		       display.show(returnValue);
		   }
               }
            }
         } else {
            if (display != null && showResult) 
               display.refresh();
         }
      } catch (Exception e) {
         Throwable te = Exceptions.getTargetException(e);
         Log.trace("threads",this+" TargetException is "+te);
         if (te instanceof TimeoutException) {
            Log.trace("threads",this+" Timeout");
            DialogView dialog = ((TimeoutException)te).getDialog();
            display.closeWindow(dialog);
            display.addTimedoutDialog(dialog);
         } else if (display != null && showResult) {
            Log.trace("threads",te);
            if (!(te instanceof org.objectweb.jac.util.VoidException))
               display.show(te);
            else
               display.refresh();
         }
      }
      Log.trace("threads","invokeThread done "+this+": "+
                object+"."+method+Arrays.asList(parameters));
   }

   Object getReturnValue() {
      return returnValue;
   }
}

class NotAvailableException extends Exception {}
