/*
  Copyright (C) 2001-2003 Laurent Martelli <laurent@aopsys.com>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package org.objectweb.jac.aspects.gui;

import org.objectweb.jac.core.Collaboration;
import org.objectweb.jac.core.Wrappee;
import org.objectweb.jac.core.rtti.ClassItem;
import org.objectweb.jac.core.rtti.ClassRepository;
import org.objectweb.jac.core.rtti.CollectionItem;
import org.objectweb.jac.core.rtti.FieldItem;
import org.objectweb.jac.core.rtti.MemberItem;
import org.objectweb.jac.core.rtti.MetaItem;
import org.objectweb.jac.core.rtti.MethodItem;
import org.objectweb.jac.core.rtti.NamingConventions;
import org.objectweb.jac.core.rtti.RttiAC;
import org.objectweb.jac.util.Classes;
import org.objectweb.jac.util.Enum;
import org.objectweb.jac.util.Log;
import org.objectweb.jac.util.Stack;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import javax.swing.event.TableModelEvent;
import javax.swing.table.AbstractTableModel;

/**
 * The data model for tables. */

public class TableModel extends AbstractTableModel
   implements ExtendedTableModel, ObjectUpdate, CollectionUpdate
{
   List rows = new Vector();
   List objects = new Vector();
   String[] headers;
   ClassItem[] classes;
   MemberItem[] members;

   CollectionItem collection;
   Object substance;

   /**
    * Creates a new table model.
    *
    * @param collection the substance collection
    * @param substance the object that holds the collection
    * @param factory the used view factory */
 
   public TableModel(CollectionItem collection, 
                     Object substance, String viewName,
                     ViewFactory factory) {
      this.collection = collection;
      this.substance = substance;

      members = GuiAC.getMembersOrder(collection,viewName);
      Log.trace("gui.generic","membersOrder : " + 
                (members!=null?(Arrays.asList(members).toString()):"[]"));
      if (members==null) {
         ClassItem clit = collection.getComponentType();
         if (clit!=null  && !(collection.isMap() && !RttiAC.isIndex(collection))) {
            Log.trace("gui.generic","class : " + clit.getName());
            if (members==null)
               members = (MemberItem[])clit.getAttribute(GuiAC.TABLE_MEMBERS_ORDER);
            Log.trace("gui.generic","class tableMembersOrder : " + 
                      (members!=null?(Arrays.asList(members)+""):""));
            if (members==null)
               members = (MemberItem[])GuiAC.getView(clit,viewName).getAttributesOrder();
            Log.trace("gui.generic","class attributesOrder : " +
                      (members!=null?(Arrays.asList(members)+""):""));
            if (members==null)
               members = clit.getFields();
         } else if (!RttiAC.isIndex(collection)) {
            members = new MemberItem[2];
            clit = ClassRepository.get().getClass(Map.Entry.class);
            members[0] = clit.getField("key");
            members[1] = clit.getField("value");
         } else {
            members = new MemberItem[0];
         }
      }
      Log.trace("gui.generic","columns : " + Arrays.asList(members));

      FieldItem oppositeRole = 
         (FieldItem)Collaboration.get().getAttribute(GuiAC.OPPOSITE_ROLE);
      if (oppositeRole instanceof CollectionItem) {
         Log.trace("associations","Ignoring collection oppositeRole "+oppositeRole);
         oppositeRole = null;
      }

      // Ensure that members contains no oppositeRole
      MemberItem[] tmp = new MemberItem[members.length];
      int j=0;
      for(int i=0; i<members.length; i++) {
         if (members[i]!=oppositeRole) {
            tmp[j] = members[i];
            j++;
         }
      }
      members = new MemberItem[j];
      System.arraycopy(tmp,0,members,0,j);

      int nbColumns = members.length;
      headers = new String[nbColumns];
      classes = new ClassItem[nbColumns];

      Stack context = GuiAC.getGraphicContext();
      for (int i=0; i<nbColumns; i++) {
         if (members[i] instanceof FieldItem) {
            headers[i] = GuiAC.getLabel(members[i],context);
         }
         classes[i] = members[i].getTypeItem();
      }

      buildData();
      Utils.registerCollection(substance,collection,this);
   }

   public CollectionItem getCollection() {
      return collection;
   }

   static class CellLocation {
      int row;
      int column;
      public CellLocation(int row, int column) {
         this.row = row;
         this.column = column;
      }
   }

   public Object getCellRenderer(View tableView, int column, 
                                 ViewFactory factory, DisplayContext context) {
      return getCellRenderer(tableView,substance,
                             members[column],headers[column],
                             factory,context);
   }

   public static Object getCellRenderer(View tableView, 
                                        Object substance,
                                        MemberItem member, 
                                        String header,
                                        ViewFactory factory, 
                                        DisplayContext context) {
      if (member instanceof FieldItem) {
         FieldItem field = (FieldItem)member;
         Enum enum = (Enum)field.getAttribute(GuiAC.FIELD_ENUM); 
         if (enum!=null) {
            return factory.createView(
               field.getName(),"cell:Enum",
               new Object[] {enum},
               context);
         } 

         Stack typeNames = new Stack();
         if (field.isReference())
            typeNames.push("cell:Reference");
         else if (field instanceof CollectionItem)
            typeNames.push("cell:List");
         typeNames.push("cell:"+field.getTypeItem().getName());
         MetaItem type = RttiAC.getFieldType(field);
         if (type!=null)
            typeNames.push("cell:"+type.getName());
         Log.trace("gui.table","types for "+field+": "+typeNames);
         while (!typeNames.empty()) {
            String typeName = (String)typeNames.pop();
            if (factory.hasViewerFor(typeName)) {
               try {
                  FieldView cellViewer = (FieldView)factory.createView(
                     "table["+field.getName()+"]",
                     typeName,
                     new Object[0],context);
                  if (cellViewer instanceof TableCellViewer) {
                     ((TableCellViewer)cellViewer).setTable(tableView);
                  }
                  Log.trace("gui.table","viewer= "+cellViewer);
                  cellViewer.setField(field);
                  return cellViewer;
               } catch (Exception e) {
                  Log.error("Failed to instanciate TableCellRenderer "+
                            field.getName()+" for column "+header);
                  e.printStackTrace();
               }
            } else {
               Log.trace("gui.table","no viewer found for type "+typeName);
            }
         }
      } else if (member instanceof MethodItem) {
         MethodItem method = (MethodItem)member;
         try {
            MethodView cellViewer = (MethodView)factory.createView(
               "table["+method.getName()+"]",
               "cell:Method",
               new Object[] {substance, method},context);
            if (cellViewer instanceof TableCellViewer) {
               ((TableCellViewer)cellViewer).setTable(tableView);
            }
            Log.trace("gui.table","viewer= "+cellViewer);
            return cellViewer;
         } catch (Exception e) {
            Log.error("Failed to instanciate TableCellRenderer "+
                      method.getName()+" for column "+header);
            e.printStackTrace();
         }
      }
      return null;
   }

   /**
    * Populate the table model with the objects from the collection.
    */
   void buildData() {
      Collection c = collection.getActualCollectionThroughAccessor(substance);
      Log.trace("gui.generic","substance : " + substance);
      Log.trace("gui.generic","objects : " + new Vector(c));
      int nbColumns = members.length;
      Iterator i = c.iterator();
      int row = 0;
      while (i.hasNext()) {
         Object obj = i.next();
         Object[] data = new Object[nbColumns];
         if (obj!=null) {
            for (int j=0; j<nbColumns; j++) {
               try {
                  if (members[j] instanceof FieldItem) { 
                  data[j] = ((FieldItem)members[j]).getThroughAccessor(obj);
                  if (data[j] instanceof Wrappee && !(data[j] instanceof Collection)) {
                     Utils.registerObject(data[j],this,new CellLocation(row,j));
                  }
                  } else if (members[j] instanceof MethodItem) {
                     data[j] = members[j].getName();
                  }
               } catch (Exception e) {
                  Log.error("Failed to build table cell for "+obj+"."+members[j].getName());
               }
            }
         }
         Log.trace("gui.generic",2,"add "+Arrays.asList(data));
         addRow(obj,data);
         row++;
      }      
   }

   public MemberItem[] getMembers() {
      return members;
   }

   public String[] getHeaders() {
      return headers;
   }

   public int getRowCount() {
      return rows.size();
   }
   public int getColumnCount() {
      return headers.length;
   }

   public String getColumnName(int column) {
      return headers[column];
   }
   public Class getColumnClass(int column) {
      Log.trace("gui.table",2,
                "getColumnClass("+column+") -> "+classes[column].getName());
      return Classes.getPrimitiveTypeWrapper(classes[column].getActualClass());
   }
   public Object getValueAt(int row, int column) {
      Log.trace("gui.table",2,"getValueAt("+row+","+column+") -> "+
                ((Object[])rows.get(row))[column]);
      return ((Object[])rows.get(row))[column];
   }
   public Object getObject(int index) {
      return objects.get(index);
   }
   public int indexOf(Object object) {
      return objects.indexOf(object);
   }
   public Object getObject(int row, int col) {
      return ((Object[])rows.get(row))[col];
   }
   public Object[] getRow(int row) {
      return (Object[])rows.get(row);
   }
   public boolean isCellEditable(int row, int column) {
      return false;
   }

   public void addRow(Object object, Object[] data) {
      objects.add(object);
      rows.add(data);
      Utils.registerObject(object,this);
   }

   // ObjectUpdate interface
   public void objectUpdated(Object substance,Object param) {
      if (param==null) {
         int index = objects.indexOf(substance);
         if (index!=-1) {
            int nbColumns = getColumnCount();
            Object[] data = new Object[nbColumns];
            for (int j=0; j<nbColumns; j++) {
               if(members[j] instanceof FieldItem) {
                  data[j] = ((FieldItem)members[j]).getThroughAccessor(substance);
               }
            }
            rows.set(index,data);
         }
         fireTableRowsUpdated(index,index);
      } else {
         CellLocation location = (CellLocation)param;
         fireTableCellUpdated(location.row,location.column);
      }
   }

   // CollectionUpdate
   public void onChange(Object substance, CollectionItem collection, 
                        Object value, Object param) {
      Log.trace("gui.events","onChange "+substance+"."+collection);
      unregisterViews();
      int numRows = objects.size()-1;
      objects.clear();
      rows.clear();
      if (numRows>0)
         fireTableRowsDeleted(0,numRows);
      buildData();
      fireTableRowsInserted(0,objects.size()-1);
   }

   public void onAdd(Object substance, CollectionItem collection, 
                     Object value, Object added, Object param) {
      Log.trace("gui.events","onAdd "+substance+"."+collection);
      // it's not that easy to optimize because we don't know the
      // position of the added object
      onChange(substance,collection,value,param);
   }

   public void onRemove(Object substance, CollectionItem collection, 
                        Object value, Object removed, Object param) {
      Log.trace("gui.events","onRemove "+substance+"."+collection);
      onChange(substance,collection,value,param);
   }

   /**
    * Register ourself as a view on all objects of the collection
    */
   protected void registerViews() {
      Log.trace("gui","TableModel.registerViews "+objects.size());
      Iterator i = objects.iterator();
      while (i.hasNext()) {
         Object object = i.next();
         Utils.registerObject(object,this);
      }
   }

   /**
    * Unregister ourself as a view on all objects of the collection
    */
   protected void unregisterViews() {
      Log.trace("gui","TableModel.unRegisterViews "+objects.size());
      Iterator i = objects.iterator();
      while (i.hasNext()) {
         Object object = i.next();
         Utils.unregisterObject(object,this);
      }      
   }

   public void close() {
      unregisterViews();
      Utils.unregisterCollection(substance,collection,this);
   }
}
