/*
  Copyright (C) 2001-2003 Laurent Martelli <laurent@aopsys.com>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package org.objectweb.jac.aspects.gui.web;

import java.io.PrintWriter;
import org.objectweb.jac.aspects.gui.*;
import org.objectweb.jac.core.NameRepository;
import org.objectweb.jac.core.rtti.ClassRepository;
import org.objectweb.jac.core.rtti.CollectionItem;
import org.objectweb.jac.core.rtti.FieldItem;
import org.objectweb.jac.util.Log;


public abstract class AbstractCollection extends AbstractView
   implements CollectionListener, CollectionView
{
   protected CollectionItem collection;
   protected Object substance; 
   protected CollectionModel model;
   protected boolean autoUpdate = true;
   protected boolean isEditor = true;

   /* index of selected row (-1 if no row if selection is empty) */
   int selected = -1;

   /* number of rows to display on each page */
   int rowsPerPage = 10;
   /* do we split the collection in multiple pages */
   boolean split;

   /* do we show a column with row numbers */
   boolean showRowNumbers;
   
   /* current offset */
   int startIndex = 0;

   protected org.objectweb.jac.aspects.gui.CollectionItemView itemView;

   public AbstractCollection(ViewFactory factory, DisplayContext context,
                             CollectionItem collection, Object substance,
                             CollectionModel model, 
                             org.objectweb.jac.aspects.gui.CollectionItemView itemView) {
      super(factory,context);

      this.collection = collection;
      this.substance = substance;
      this.model = model;
      this.rowsPerPage = GuiAC.getNumRowsPerPage(collection);
      this.startIndex = GuiAC.getStartIndex(collection);
      this.showRowNumbers = GuiAC.isShowRowNumbers(collection);
      this.itemView = itemView;
      split = rowsPerPage>0;
      sort();
   }

   public AbstractCollection() {
   }

   public void setSubstance(Object substance) {
      this.substance = substance;
   }

   public CollectionModel getCollectionModel() {
      return model;
   }

   public boolean isEditor() {
      return isEditor;
   }

   public void setEditor(boolean isEditor) {
      this.isEditor = isEditor;
   }

   public void setAutoUpdate(boolean autoUpdate) {
      this.autoUpdate = autoUpdate;
   }

   /**
    * Generate HTML code for a collection's event (add, remove, ...)
    * @param out write HTML to this writer
    * @param event name of the associated event
    * @param icon name of icon resource to display
    * @param label text to display
    */
   void genCollectionEvent(PrintWriter out,
                           String event,
                           String icon,
                           String label) {

      JacRequest request = ((WebDisplay)context.getDisplay()).getRequest();
      if (request.isIEUserAgent()) {
         out.println(
            "<table class=\"method\"><td>"+
            iconElement(ResourceManager.getResource(icon),"")+
            eventURL(label,event,"").toString()+
            "</td></table>");
      } else {
         out.println(
            "<span class=\"method\">"+
            iconElement(ResourceManager.getResource(icon),"")+
            eventURL(label,event,"").toString()+
            "</span>");
      }
   }

   /**
    * Sorts the collection with the column index stored in the context
    * if any.
    */
   public abstract void sort();

   public void onView(Object object) {
      Log.trace("gui.events",toString()+".onView("+object+")");
      selected = model.indexOf(object);

      // used for CollectionItemView
      CollectionPosition extra = new CollectionPosition(
         this,
         collection,selected,substance);

      EventHandler.get().onSelection(context,collection,object,
                                     null,extra,true);
   }

   // HTMLViewer interface

   /**
    * Generate HTML for adder, remover, and prev/next buttons
    * @param out print HTML code to this writer 
    */
   protected void genAdderAndRemover(PrintWriter out) {
      if (collection.getSubstance(substance)==null)
         return;

      boolean addable = GuiAC.isAddable(substance,collection);
      boolean removable = GuiAC.isRemovable(substance,collection);
      if (!(addable ||
            (model.getRowCount()>0 && removable) ||
            (startIndex>0) ||
            (split && startIndex+rowsPerPage<model.getRowCount()))) {
         return;
      }
      out.println("<div class=\"methods\">");
      if (addable && isEditor) {
         if (GuiAC.isAutoCreate(collection)) {
            genCollectionEvent(out,"onAddToCollection","new_icon",GuiAC.getLabelAdd());
            /*
            genCollectionEvent(out,"onAddToCollection","new_icon","add new");
            genCollectionEvent(out,"onAddExistingToCollection","new_icon","add existing");
            */
         } else {
            genCollectionEvent(out,"onAddToCollection","new_icon",GuiAC.getLabelAdd());
         }
      }
      if (startIndex>0) {
         out.println("<a href=\""+eventURL("onFirst")+"\">"+
                     iconElement(ResourceManager.getResource("first_icon"),"first")+
                     "</a>");
         out.println("<a href=\""+eventURL("onPrevious")+"\">"+
                     iconElement(ResourceManager.getResource("previous_icon"),"previous")+
                     "</a>");
      }
      if (rowsPerPage>0 && model.getRowCount()>0 && 
          (startIndex>0 || startIndex+rowsPerPage<model.getRowCount())) {
         out.println("["+(startIndex+1)+"-"+
                     Math.min(startIndex+rowsPerPage,model.getRowCount())+"/"+
                     model.getRowCount()+"]");
      }
      if (split && startIndex+rowsPerPage<model.getRowCount()) {
         out.println("<a href=\""+eventURL("onNext")+"\">"+
                     iconElement(ResourceManager.getResource("next_icon"),"next")+
                     "</a></span>");
         out.println("<a href=\""+eventURL("onLast")+"\">"+
                     iconElement(ResourceManager.getResource("last_icon"),"last")+
                     "</a></span>");
      }
      if (split && model.getRowCount()>rowsPerPage) {         
         out.println("["+(startIndex/rowsPerPage+1)+"/"+
                     (int)Math.ceil((float)model.getRowCount()/(float)rowsPerPage)+"]");
      }

      out.println("</div>");
   }

   // CollectionView interface

   public void setSelected(int index) {
      selected = index;      
   }

   public void setField(FieldItem field) {
      collection = (CollectionItem)field;
   }

   public FieldItem getField() {
      return collection;
   }

   public void setValue(Object value) {
   }

   public void updateModel(Object substance) {
   }

   // CollectionListener interface

   public void onView(int index) {
      Log.trace("gui.events",toString()+".onView("+index+")");
      selected = index;

      // used for CollectionItemView
      CollectionPosition extra = new CollectionPosition(
         this,
         collection,index,substance);

      EventHandler.get().onSelection(context,collection,model.getObject(index),
                                     null,extra,true);
   }

   public void onViewObject(String name) {
      Log.trace("gui.events",toString()+".onViewObject("+name+")");
      Object object = NameRepository.get().getObject(name);
      EventHandler.get().onSelection(context,collection,object,
                                     null,null,true);
   }

   public void onRemove(int index) {
      Log.trace("gui.events",toString()+".onRemove("+index+")");
      Object toRemove = model.getObject(index);
      
      EventHandler.get()
         .onRemoveFromCollection(
            context, collection.getSubstance(substance), collection, toRemove, false);
   }

   public void onTableInvoke(int index,String methodName) {
      Log.trace("gui.events",toString()+".onTableInvoke("+index+")");
      selected = index;
      EventHandler.get().onInvoke(
         context,model.getObject(index),
         ClassRepository.get().getClass(model.getObject(index))
         .getMethod(methodName));
   }
   
   public void onAddToCollection() {
      EventHandler.get().onAddToCollection(
         context,collection.getSubstance(substance),collection,false);
      sort();
   }

   public void onAddExistingToCollection() {
      EventHandler.get().onAddToCollection(
         context,collection.getSubstance(substance),collection,true);
      sort();
   }

   public void onRemoveFromCollection() {
      EventHandler.get().onRemoveFromCollection(
         context,collection.getSubstance(substance),collection,true);
      sort();
   }

   public void onNext() {
      if (startIndex+rowsPerPage<model.getRowCount())
         startIndex += rowsPerPage;
      context.getDisplay().refresh();
   }

   public void onLast() {
      startIndex = model.getRowCount() - model.getRowCount()%rowsPerPage;
      if (startIndex==model.getRowCount())
         startIndex -= rowsPerPage;
      context.getDisplay().refresh();
   }

   public void onPrevious() {
      startIndex -= rowsPerPage;
      if (startIndex<0)
         startIndex = 0;
      context.getDisplay().refresh();
   }

   public void onFirst() {
      startIndex = 0;
      context.getDisplay().refresh();
   }

   /**
    * Returns an HTML link to remove the element at a given position.
    */
   protected String removeLink(int position) {
      // We should use eventURL
      return "<a href=\""+eventURL("onRemove")+
         "&amp;index="+position+"\">"+
         iconElement(ResourceManager.getResource("remove_icon"),"remove")+"</a>";
   }

   protected String viewLink(int position) {
      // We should use eventURL
      return "<a href=\""+eventURL("onView")+
         "&amp;index="+position+"\">"+
         iconElement(ResourceManager.getResource("view_icon"),"view")+"</a>";
   }

   /**
    * Build an HTML link with an image showing if a column is used to
    * sort the collection
    * @param column the index of the column 
    * @param text additional text to put in the link
    * @return the HTML code of the link
    */
   protected String sortLink(int column,String text) {
      SortCriteria criteria = ((TableSorter)model).getSortCriteria(column);
      String iconResource;
      if (criteria==null) {
         iconResource = "small_down_icon";
      } else {
         if (criteria.isAscending()) 
            iconResource = "small_down_icon_selected";
         else 
            iconResource = "small_up_icon_selected";
      }
      return 
         "<a href=\""+eventURL("onHeaderClick")+"&amp;col="+column+"\""
         + " title=\"sort\">"
         + iconElement(ResourceManager.getResource(iconResource),"sort")
         + " " + text
         + "</a>";
   }

   // When disabled == true, readValue() does nothing
   boolean disabled = false;
   public boolean isEnabled() {
      return !disabled;
   }
   public void setEnabled(boolean enabled) {
      Log.trace("gui.editor",(enabled?"enable ":"disabled")+this);
      this.disabled = !enabled;
   }
}
