/*
  Copyright (C) 2002-2003 Renaud Pawlak <renaud@aopsys.com>, 
                          Laurent Martelli <laurent@aopsys.com>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package org.objectweb.jac.aspects.gui.web;

import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import javax.servlet.http.HttpServletResponse;
import org.mortbay.http.HttpContext;
import org.mortbay.http.HttpServer;
import org.mortbay.http.handler.ResourceHandler;
import org.mortbay.jetty.servlet.ServletHandler;
import org.mortbay.util.InetAddrPort;
import org.mortbay.util.MultiException;
import org.objectweb.jac.aspects.gui.*;
import org.objectweb.jac.aspects.session.SessionAC;
import org.objectweb.jac.core.Collaboration;
import org.objectweb.jac.core.Jac;
import org.objectweb.jac.core.NameRepository;
import org.objectweb.jac.core.rtti.AbstractMethodItem;
import org.objectweb.jac.lib.Attachment;
import org.objectweb.jac.util.ExtArrays;
import org.objectweb.jac.util.ExtBoolean;
import org.objectweb.jac.util.Log;
import org.objectweb.jac.util.Strings;


/**
 * This class provides a server for web clients using a thin client
 * protocol and implements the <code>Display</code> interface to
 * assume data inputs and outputs between JAC objects and web clients.
 *
 * <p>This inferface is typically used by Java web clients and more
 * specifically by servlets. The implementation provided by JAC is
 * <code>JacServlet</code>.
 *
 * @author <a href="mailto:renaud@cnam.fr">Renaud Pawlak</a>
 * @author <a href="mailto:laurent@aopsys.com">Laurent Martelli</a> 
 */

public class WebDisplay implements CustomizedDisplay {
   ViewFactory factory;
   // customizedID -> customized
   Hashtable frames = new Hashtable();

   /* The session */
   Session session;

   /**
    * Create a new web display
    * @param factory the ViewFactory of the display
    * @param sessionId the sessionId of the display
    */
   public WebDisplay(ViewFactory factory, String sessionId) {
      this.factory = factory;
      session = new Session(sessionId);
      Collaboration.get().addAttribute( 
         SessionAC.SESSION_ID, sessionId );
   }

   public static final String RESPONSE = "gui.web.response";
   public static final String REQUEST = "gui.web.request";

   /**
    * Set the servlet response for the current collaboration
    * @param response the servlet response object
    */
   public static void setResponse(HttpServletResponse response) {
      Collaboration.get().addAttribute(RESPONSE,response);
   }
   /**
    * Returnsx the servlet response of the current collaboration
    * @return the servlet response of the current collaboration
    */
   public static HttpServletResponse getResponse() {
      return (HttpServletResponse)Collaboration.get().getAttribute(RESPONSE);
   }

   /**
    * Sets the current JacRequest in the context.
    * @param request the JacRequets
    */ 
   public static void setRequest(JacRequest request) {
      Log.trace("web.request",
                "Setting request for "+Strings.hex(Collaboration.get())+
                ": "+request);
      Collaboration.get().addAttribute(REQUEST,request);
   }

   /**
    * Returns the current JacRequest contained in the context.
    * @return the current JacRequest contained in the context
    */
   public static JacRequest getRequest() {
      return (JacRequest)Collaboration.get().getAttribute(REQUEST);
   }

   /**
    * Read values of field editors from http request
    * @param editors field editors for which to read the value
    * @param request the http request where to read values from
    * @param commit if true, commit() will be called on the field
    * editors after readValue() 
    * @see HTMLEditor#readValue(Object)
    * @see HTMLEditor#commit()
    */
   public static void readValues(DisplayContext editors, JacRequest request, 
                                 boolean commit) 
   {
      Log.trace("gui.editor","reading value from context "+editors);
      Iterator i = editors.getEditors().iterator();
      while (i.hasNext()) {
         View editor = (View)i.next();
         Object parameter = request.getParameter(editor.getLabel());
         Log.trace("gui.editor","reading value for "+editor.getLabel());
         if (editor instanceof JacRequestReader) {
            ((JacRequestReader)editor).readValue(request);
         } else {
            if (((HTMLEditor)editor).readValue(parameter) && commit) {
               ((HTMLEditor)editor).commit();
            }
         }
      }      
   }

   public void fullRefresh() {
      Iterator it=frames.entrySet().iterator();
      while(it.hasNext()) {
         Map.Entry entry=(Map.Entry)it.next();
         View frame=(View)entry.getValue();
         CustomizedGUI customized=((CustomizedView)frame).getCustomizedGUI();
         frame.close();
         View newframe = factory.createView(
            customized.getTitle(),"Customized",
            new Object[] {customized,null},
            new DisplayContext(this,null));
         Log.trace("display","frame created "+newframe);
         frames.put(entry.getKey(),newframe);
         session.newRequest(new Request(newframe));
      }
   }

   public void showCustomized(String id, Object object) {
      showCustomized(id,object,null);
   }

   public void showCustomized(String id, Object object, Map panels) {
      Log.trace("display","showCustomized("+id+","+object+")");
      try {
         CustomizedGUI customized = (CustomizedGUI)object;
         if (frames.get(id)!=null) {
            if (panels!=null) {
               CustomizedView frame = (CustomizedView)frames.get(id);
               GenericFactory.initCustomized(frame.getFactory(), 
                                             frame.getContext(), 
                                             frame.getPanelView(), 
                                             customized, panels);         
            }
            refresh();
         } else {
            HTMLViewer frame = (HTMLViewer)factory.createView(
               customized.getTitle(),"Customized",
               new Object[] {customized,panels},
               new DisplayContext(this,null));
            Log.trace("display","frame created");
            frames.put(id,frame);
            session.newRequest(new Request((View)frame));
            refresh();
         }
      } catch(Exception e) {
         e.printStackTrace();
         showError("showCustomized error",e.toString());
      }
   }

   public void show(Object object) {
      Log.trace("display","show("+object+")");
      show(object,false);
   }

   protected void show(Object object, boolean newWindow) {
      if (object==null) {
         return;
         /*
      } else if (object instanceof Throwable) {
         showError("Exception",object.toString());
         ((Throwable)object).printStackTrace();
         */
      } else if (object instanceof InputStream) {
         try {
            byte[] buffer = new byte[4096];
            InputStream input = (InputStream)object;
            int length;
            while ((length=input.read(buffer))!=-1) {
               getResponse().getOutputStream().write(buffer,0,length);
            }
         } catch (IOException e) {
            Log.error("Failed to output stream: "+e);
            e.printStackTrace();
         }
      } else if (object instanceof Attachment) {
         try {
            DisplayContext context = new DisplayContext(this,null);
            HttpServletResponse response = getResponse();
            response.setContentType(((Attachment)object).getMimeType());
            response.getOutputStream().write(((Attachment)object).getData());
         } catch (IOException e) {
            Log.error("Failed to output stream: "+e);
            e.printStackTrace();
         } finally {
            getRequest().setResponse();
         }
      } else {
         DisplayContext context = new DisplayContext(this,null);
         View objectView = factory.createObjectView("object",object,context);

         String title;
         if (object!=null) {
            Class substance_type = object.getClass();
            String tn = substance_type.getName();
            
            title = tn.substring( tn.lastIndexOf('.') + 1) + " " +
               GuiAC.toString(object);
         } else {
            title= "<null>";
         }

         View page = factory.createView(
            title,"Window",
            new Object[] {objectView,ExtBoolean.valueOf(newWindow)},
            context);
         session.newRequest(new Request(page));
         refresh();
      }
   }

   public void openView(Object object) {
      Log.trace("display","openView("+object+")");
      show(object,true);
   }

   public void newObject(Object object) {
      Log.trace("display","newObject("+object+")");
   }

   public boolean showModal(Object object, String title, String header, 
                            Object parent,
                            boolean okButton, boolean cancelButton, 
                            boolean closeButton ) {
      Log.trace("display","showModal("+object+","+title+")");
      DisplayContext context = new DisplayContext(this,null);
      Log.trace("display","new context "+context);
      View objectView = factory.createObjectView("object",object,context);

      return showModal(objectView,context,
                       title,header,parent,
                       okButton,cancelButton,closeButton);
   }

   public boolean showModal(Object object, 
                            String viewType, String[] viewParams,
                            String title, String header, 
                            Object parent,
                            boolean okButton, boolean cancelButton, 
                            boolean closeButton ) {
      Log.trace("display","showModal("+object+","+title+","+viewType+")");
      DisplayContext context = new DisplayContext(this,null);
      View objectView = factory.createView("object",viewType,
                                           ExtArrays.add(object,viewParams),context);

      return showModal(objectView,context,
                       title,header,parent,
                       okButton,cancelButton,closeButton);
   }

   public boolean showModal(View objectView, DisplayContext context,
                            String title, String header, 
                            Object parent,
                            boolean okButton, boolean cancelButton, 
                            boolean closeButton ) {
      Log.trace("display","objectView = "+objectView);
      if (objectView instanceof org.objectweb.jac.aspects.gui.EditorContainer)
         ((org.objectweb.jac.aspects.gui.EditorContainer)objectView).setShowButtons(false);
         
      DialogView page = 
         (DialogView)factory.createView(
            title,"Dialog",
            new Object[] {objectView,parent,title,header},
            context);
      context.setWindow(page);
      session.newRequest(new Request(page));
      refresh();

      try {
         boolean result = page.waitForClose();
         WebDisplay display = (WebDisplay)context.getDisplay();
         display.setResponse(((Dialog)page).getResponse());
         display.setRequest(((Dialog)page).getRequest());
         return result;
      } catch (TimeoutException timeout) {
         addTimedoutDialog(page);
         throw timeout;
      }
   }

   public boolean showInput(Object substance, AbstractMethodItem method, 
                            Object[] parameters) {
      Log.trace("display","showInput("+substance+","+method+","+
                Arrays.asList(parameters)+")");
      DisplayContext dc = (DisplayContext)Collaboration.get()
         .getAttribute(GuiAC.DISPLAY_CONTEXT);
      //if (dc==null) {
         dc = new DisplayContext(this,null);
         //}
      DialogView page = GenericFactory.createInputDialog(substance,
                                                         method,parameters,dc);
      dc.setWindow(page);
      session.newRequest(new Request(page));
      refresh();
      try {
         if (page.waitForClose()) {
            EditorContainer inputView = (EditorContainer)page.getContentView();
            Iterator it = inputView.getEditors().iterator();
            int i=0;
            JacRequest request = getRequest();
            while (it.hasNext()) {
               FieldEditor editor = (FieldEditor)it.next();
               parameters[i] = editor.getValue();
               i++;
            }
            setResponse(((Dialog)page).getResponse());
            setRequest(((Dialog)page).getRequest());
            return true;
         } else {
            setResponse(((Dialog)page).getResponse());
            setRequest(((Dialog)page).getRequest());
            return false;
         }
      } catch (TimeoutException timeout) {
         addTimedoutDialog(page);
         throw timeout;
      }
   }

   public boolean showMessage(String message, String title,
                              boolean okButton, 
                              boolean cancelButton, 
                              boolean closeButton ) {
      Log.trace("display","showMessage("+message+","+title+")");
      return showModal(null,title,message,null,okButton,cancelButton,closeButton);
   }

   public void showMessage(String title, String message) {
      try {
         Log.trace("display","showMessage("+title+","+message+")");
         DisplayContext context=(DisplayContext)Collaboration.get()
            .getAttribute(GuiAC.DISPLAY_CONTEXT);
         if(context==null) {
            context=new DisplayContext(this,null);
         }
         View label = factory.createView("error message","text",
                                         new Object[] {message,null,null},context);
         View page = factory.createView("Object view","Window",
                                        new Object[] {label,Boolean.FALSE},context);
         session.newRequest(new Request(page));
      } finally {
         refresh();
      }
   }

   public void showError(String title, String message) {
      showMessage(title,message);
   }

   public void showStatus(String message) {
   }

   public void applicationStarted() {
   }
   

   public CustomizedView getCustomizedView(String customizedID) {
      return (CustomizedView)frames.get(customizedID);
   }
   public Collection getCustomizedViews() {
      return frames.values();
   }
   public ViewFactory getFactory() {
      return factory;
   }

   String displayID;

   public String getDisplayID() {
      return displayID;
   }

   public void setDisplayID(String displayID) {
      this.displayID = displayID;
   }

   public void close() {
      // close all customized guis
      Iterator i = frames.values().iterator();
      while(i.hasNext()) {
         View view = (View)i.next();
         view.close();
      }
   }   

   // definition of AbstractServer abstract methods

   public Object[] buildParameterValues(AbstractMethodItem method, String[] params) {
      Object[] result = new Object[params.length];
      Class[] pts = method.getParameterTypes();
      for(int i=0; i<params.length; i++)
      {
         //         HTMLEditor editor = getEditorComponent(null,method,i);
         //         result[i] = editor.getValue(params[i]);
      }
      return result;
   }

   public void refresh() {
      Log.trace("display","refresh");
      Object view = session.getCurrentRequest().getView();
      if (view == null) {
         showError("Display error","Nothing to refresh yet");
         Log.trace("web","Nothing to refresh yet");
      } else {
         try {
            Log.trace("web",
                      "genHTML on "+view+"("+Strings.hex(getResponse())+")");
            ((HTMLViewer)view).genHTML(getResponse().getWriter());
         } catch(Exception e) {
            Log.error("refresh failed: "+e);
            e.printStackTrace();
         } finally {
            getRequest().setResponse();
         } 
         Log.trace("web","refresh DONE...");
      }      
   }

   static HttpServer httpServer;
   public static HttpServer getHttpServer() {
      if (httpServer==null) {
         httpServer = new HttpServer();
      }
      return httpServer;
   } 

   /** Alreeady started GUIs */
   static HashSet startedGuis = new HashSet();
   /** TCP Ports the servlet engine listens to */
   static HashSet openedPorts = new HashSet();

   /**
    * Start JAC's internal webserver. You can then interact with the
    * application through an URL like this: 
    * http://<hostname>:<port>/<application>/<guiID>
    *
    * @param application name of the application
    * @param guiIDs name of windows. gui_name[:port]
    * @param defaultPort default TCP port on which to listen */
   public static void startWebServer(String application, String[] guiIDs, int defaultPort) 
      throws IOException, MultiException
   {
      Log.trace("web","startWebServer on port "+defaultPort);
      httpServer = getHttpServer();
      HttpContext context = httpServer.getContext("/jac");

      ServletHandler servletHandler = new ServletHandler();
      for (int i=0; i<guiIDs.length;i++) {
         if (startedGuis.contains(guiIDs[i])) {
            Log.trace("web","skipping already registered GUI "+guiIDs[i]);
         } else {
            Log.trace("web","register GUI "+guiIDs[i]);

            String split[] = Strings.split(guiIDs[i],":");
            String gui = split[0];
            int port = split.length>=2 ? Integer.parseInt(split[1]) : defaultPort;
            
            if (!openedPorts.contains(new Integer(port))) {
               httpServer.addListener(new InetAddrPort(port));
               openedPorts.add(new Integer(port));
            }
            
            // Make sure the servlet will have the JAC ClassLoader
            // or it won't get the right NameRepository
            context.setClassLoader(WebDisplay.class.getClassLoader());
         
            servletHandler.addServlet("Jac","/"+gui,
                                      "org.objectweb.jac.aspects.gui.web.JacLocalServlet");
         }
      }
      context.addHandler(servletHandler);

      context = httpServer.getContext("/jac/resources");
      context.setBaseResource(new ClasspathResource());
      context.setClassLoader(WebDisplay.class.getClassLoader());
      context.addHandler(new ResourceHandler());

      startWebServer();
   }

   /**
    * Start the web server if it is not already started
    */
   public static synchronized void startWebServer() throws MultiException {
      if (httpServer.isStarted()) {
         try {
            httpServer.stop();
         } catch (Exception e) {
            e.printStackTrace();
         }
      } 
      httpServer.start();
   }

   // View -> (String)id
   Hashtable ids = new Hashtable();
   // (String)id -> View
   Hashtable views = new Hashtable();

   /**
    * Compute an id for a view and register it.
    * @param view the view to register
    * @return the id of the view
    * @see #unregisterView(View)
    * @see #getView(String)
    */
   public String registerView(View view) {
      String id = (String)ids.get(view);
      if (id==null) {
         id = Integer.toString(view.hashCode());
         ids.put(view,id);
         views.put(id,view);
         Log.trace("web",2,toString()+".registerView("+view+") -> "+id);
      }
      return id;
   }

   /**
    * Unregister a view.
    * @param view the view to register
    * @see #registerView(View)
    */
   public void unregisterView(View view) { 
      Object id = ids.get(view);
      Log.trace("web",2,toString()+".unregisterView("+view+") "+id);
      ids.remove(view);
      if (id!=null) {
         views.remove(id);
      }
   }

   /**
    * Returns the view registered with a given id
    * @param id the id of the view
    * @return the registered view with that id, null if there is none
    * @see #registerView(View)
    */
   public View getView(String id) {
      Log.trace("web",2,toString()+".getView("+id+")");
      return (View)views.get(id);
   }

   /**
    * Gets the ID of a registered view.
    *
    * @param view the view
    * @return the ID of the view 
    */
   public String getViewID(View view) {
      return (String)ids.get(view);
   }

   // Strings
   HashSet timedoutDialogs = new HashSet();
   public void addTimedoutDialog(DialogView dialog) {
      timedoutDialogs.add(getViewID(dialog));
   }
   /**
    * Tells wether a view ID corresponds to a timedout dialog
    */
   public boolean isTimedout(String viewID) {
      return timedoutDialogs.contains(viewID);
   }

   String servletName;
   /**
    * Set the name of the servlet
    * @param name the name of the servlet
    */
   public void setServletName(String name) {
      servletName = name;
   }
   /**
    * Returns the name of the servlet
    */
   public String getServletName() {
      return servletName;
   }

   public void closeWindow(View window) {
      Log.trace("web","closeWindow "+window);
      Iterator i = session.getRequests().iterator();
      Request request = null;
      while (i.hasNext()) {
         request = (Request)i.next();
         if (request.getView().equals(window)) {
            // we MUST do the close() before remove() because if
            // close() fails with an exception, we do not want to
            // remove.
            ((View)request.getView()).close();
            session.getRequests().remove(request);
            break;
         }
      }
   }

   public Session getSession() {
      return session;
   }
}
