/*
  Copyright (C) 2001-2003 Laurent Martelli <laurent@aopsys.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package org.objectweb.jac.aspects.persistence;

import java.util.*;
import org.objectweb.jac.core.AspectComponent;
import org.objectweb.jac.core.Interaction;
import org.objectweb.jac.core.Wrappee;
import org.objectweb.jac.core.Wrapping;
import org.objectweb.jac.util.ExtBoolean;
import org.objectweb.jac.util.Log;

/**
 * A wrapper for the Map interface.
 */
public class MapWrapper extends CollectionWrapper {

   public MapWrapper(AspectComponent ac, boolean isLoaded) {
      super(ac,isLoaded);
   }

   protected void doLoad(Wrappee wrappee) throws Exception {
      OID oid = getOID(wrappee);
      Map map = getStorage().getMap(oid);
      // Keep the number of dynamic invocations to the minimum
      HashMap normalizedMap = new HashMap(map.size());
      Iterator i = map.entrySet().iterator();
      while (i.hasNext()) {
         Map.Entry entry = (Map.Entry)i.next();
         try {
            normalizedMap.put(normalizeOutput(entry.getKey()),
                              normalizeOutput(entry.getValue()));
         } catch (NoSuchOIDError e) {
            Log.error("MapWrapper.doLoad("+oid+
                      "): skipping entry with unknown OID "+entry);
         }
      }
      attrdef(ATTR_ADDED,"true");
      Wrapping.invokeOrg(wrappee,"putAll",new Object[] { normalizedMap } );
      attrdef(ATTR_ADDED,null);
   }

   public Object containsKey(Interaction interaction) throws Exception {
      if (isLoaded) {
         return proceed(interaction);
      } else {
         return ExtBoolean.valueOf(getStorage().mapContainsKey(
            getOID(interaction.wrappee),
            normalizeInput(interaction.args[0])));
      }
   }
   
   public Object containsValue(Interaction interaction) throws Exception {
      if (isLoaded) {
         return proceed(interaction);
      } else {
         return ExtBoolean.valueOf(getStorage().mapContainsValue(
            getOID(interaction.wrappee),
            normalizeInput(interaction.args[0])));
      }
   }
   
   public Object put(Interaction interaction) throws Exception {
      Object ret = null;
      if (isLoaded)
         proceed(interaction);
      ret = normalizeOutput( 
         getStorage().putInMap(
            getOID(interaction.wrappee),
            normalizeInput(interaction.args[0]),
            normalizeInput(interaction.args[1])));
      return ret;
   }

   public Object get(Interaction interaction) throws Exception {
      if (isLoaded)
         return proceed(interaction);
      else
         return normalizeOutput(
            getStorage().getFromMap(
               getOID(interaction.wrappee),
               normalizeInput(interaction.args[0])));
   }

   public Object remove(Interaction interaction) throws Exception {
      Object result1 = null;
      boolean proceeded = false;
      if (isLoaded) {
         result1 = proceed(interaction);
         proceeded = true;
      }
      Object result2 = 
         getStorage().removeFromMap(
            getOID(interaction.wrappee),
            normalizeInput(interaction.args[0]));
      if (proceeded)
         return result1;
      else 
         return normalizeOutput(result2);
   }

   /**
    * Remove all instances from the collection
    */
   public Object clear(Interaction interaction) throws Exception {
      Object result = proceed(interaction);
      getStorage().clearMap(getOID(interaction.wrappee));
      return result;
   }

   protected long getCollectionSize(OID oid) throws Exception {
      return getStorage().getMapSize(oid);
   }

   public Object iterator(Interaction interaction) {
      return new MapIterator(interaction.wrappee);
   }
}
