/*
  Copyright (C) 2002-2003 Laurent Martelli <laurent@aopsys.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package org.objectweb.jac.aspects.persistence;

import org.objectweb.jac.aspects.naming.*;
import org.objectweb.jac.core.*;
import org.objectweb.jac.core.rtti.*;
import org.objectweb.jac.util.*;
import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * This wrapper defines persistence extensions for objects that a
 * defined persitent by a persistent aspect component. */

public class PersistenceWrapper extends Wrapper {

   public PersistenceWrapper(AspectComponent ac) {
      super(ac);
   }

   public static final String ATTR_ADDED = "persistence.added";

   // names of already loaded collections
   private HashSet loadedVectors = new HashSet();
   // names of already loaded references
   private HashSet loadedReferences = new HashSet();
   // oids of not yet loaded references (FieldItem -> oid)
   private HashMap notloadedReferences = new HashMap();

   static boolean isWrapped(Wrappee wrappee,FieldItem field) {
      Object value = field.get(wrappee);
      if (value instanceof Wrappee && 
         Wrapping.isExtendedBy((Wrappee)value,
                               null,
                               CollectionWrapper.class)) {
         Log.trace("persistence",2,field+" is wrapped"); 
         return true;
      } else {
         Log.trace("persistence",2,field+" is not wrapped"); 
         return false;
      }
   }

   /**
    * Gets the object in memory object for a "storage" value.
    * @param value the "storage" value
    * @return if value is an OID, a reference to the object with that
    * OID is returned, otherwise, value is returned.
    * @see normalizeInput(Object)
    */
   public Object normalizeOutput(Object value) throws Exception {
      if (value instanceof OID) {
         value = getAC().getObject((OID)value,null);
      }
      return value;
   }

   /**
    * Performs various stuff before storing a value in a storage. If
    * the value is a Wrappee, it is made persistent.
    * @param value the value to be stored
    * @return if value is a wrappee, the OID of the object is
    * returned, otherwise, value is returned.
    * @see normalizeInput(Object) 
    */
   public Object normalizeInput(Object value) throws Exception {
      if (value!=null) {
         Log.trace("persistence",2,
                   "normalizeInput("+value.getClass().getName()+")");
      }
      if (value instanceof Wrappee) {
         Wrappee wrappee = (Wrappee)value; // added object
         try {
            Wrapping.invokeRoleMethod(wrappee,"makePersistent",new Object[0]);
         } catch(Exception e) {
            e.printStackTrace();
         }
         
         //PersistenceWrapper wrapper = getAC().makePersistent(wrappee);
         
         if (getOID(wrappee)==null) {
            // this should never happen
            throw new InvalidOidException("oid is NULL,");
         }
         value = getOID(wrappee);
      } else {
         if (value!=null) {
            Log.trace("persistence",2,value.getClass().getName()+
                      " is not a wrappee");
         }
      }
      return value;
   }

   public OID getOID(Wrappee wrappee) {
      return getAC().getOID(wrappee);
   }

   void setOID(Wrappee wrappee,OID oid) {
      getAC().registerObject(oid,wrappee);
   }

   /**
    * Tells if the current object is persistent (role method).
    *
    * @return true if persistent */

   public boolean isPersistent(Wrappee wrappee) {
      return getOID(wrappee) != null;
   }

   /**
    * This wrapping method is called on static objects constructors as
    * defined by the pointcut of <code>PersistenceAC</code>.
    *
    * @see PersistenceAC#registerStatics */ 

   public Object handleStatic(Interaction interaction) {
      Log.trace("persistence",
                "handle static "+interaction.wrappee.getClass().getName());
      Log.trace("persistence.statics",
                "handle static "+interaction.wrappee.getClass().getName());
      Object ret = interaction.proceed();
      //wrap(null,true);
      try {
         String name = NameRepository.get().getName(interaction.wrappee);
         ClassItem cli = ClassRepository.get().getClass(interaction.wrappee);
         OID oid = getStorage().getOIDFromName(name);
         if (oid == null) {
            oid=getStorage().createObject(cli.getName());
            setOID(interaction.wrappee,oid);
            getStorage().bindOIDToName(oid,name);
            Log.trace("persistence",
                      interaction.wrappee.getClass().getName()+" isNew");
            CollectionItem[] array = new CollectionItem[0];
            wrapCollections(interaction.wrappee,true);
            initAllFields(interaction.wrappee);
         } else {
            setOID(interaction.wrappee,oid);
            // Call loadAllFields before wrapCollections so that it
            // preloads the OID of collections
            loadAllFields(interaction.wrappee);
            wrapCollections(interaction.wrappee,false);
         }
      } catch(Exception e) {
         e.printStackTrace();
      }
      return ret;
   }

   /**  
    * Makes an object persistent if it is not already. Is is assigned
    * an OID, and its members are saved in the storage.
    * @param wrappee the object to make persistent 
    */
   public void makePersistent(Wrappee wrappee) {
      try {
         Log.trace("persistence","makePersistent("+wrappee+")");
         if (isPersistent(wrappee)) {
            Log.trace("persistence",wrappee+" is already persistent");
            return;
         }
         ClassItem cli = ClassRepository.get().getClass(wrappee);
         OID oid = getStorage().createObject(cli.getName());
         setOID(wrappee,oid);
         getStorage().bindOIDToName(
            oid,
            NameRepository.get().getName(wrappee));
         Log.trace("persistence",wrappee+" isNew");

         // Initialize loadedVectors and loadedReferences.
         // It's important to do this as early as possible, in case
         // someone accesses them.

         FieldItem[] refs = cli.getReferences();
         for (int i=0; i<refs.length; i++) {
            loadedReferences.add(refs[i].getName());
         }
         CollectionItem[] colls = cli.getCollections();
         for (int i=0; i<colls.length; i++) {
            loadedVectors.add(colls[i].getName());
         }

         wrapCollections(wrappee,true);
         initAllFields(wrappee);
      } catch(Exception e) {
         Log.error("Failed to makePersistent("+wrappee+")");
         e.printStackTrace();
      }
   }

   /**
    * Load all the fields of an object from the storage and initialize
    * the object.
    *
    * @param wrappee the object
    * @param oid the object's oid
    */

   public void loadAllFields(Wrappee wrappee) throws Exception
   {
      //      checkOid();
      String lClassID = wrappee.getClass().getName();
      OID oid = getOID(wrappee);
      Log.trace("persistence","loadAllFields("+oid+")");

      // Load all fields from the storage
      ClassItem cli = ClassRepository.get().getClass(lClassID);

      // this should be faster because only one call to the storage is done
      FieldItem[] fields = cli.getFields();
      if (fields.length>0) {
         StorageField[] values = getStorage().getFields(oid,cli,fields);
         for (int i=0; i<values.length; i++) {
            if (values[i]!=null) {
               FieldItem field = values[i].fieldID;
               if (field.getActualField()!=null) {
                  try {
                     boolean force = PersistenceAC.isFieldPreloaded(field);
                     if (field.isPrimitive() && values[i].value!=null) {
                        //
                        // TODO: rajouter le preload pour les primitives ici
                        //
                        //if (force) {
                        field.set(wrappee,values[i].value);
                     } else if (field.isReference()) {
                        if (! field.hasAccessingMethods() ) {
                           // load a reference if it has no getter
                           getReference(wrappee,field);
                        } else {
                           // test if reference must be loaded here
                           // instead of being loaded when accessed
                           if (force) {
                              getReference(wrappee,field);
                           } else {
                              Log.trace("persistence","storing OID for reference "+
                                        field+"="+values[i].value);
                              notloadedReferences.put(field,values[i].value);
                           }
                        }
                     } else if (field instanceof CollectionItem) {
                        if ( ! ((CollectionItem)field).isWrappable(wrappee) ) {
                           Log.trace("persistence",
                                     "loadAllFields -> invoking getCollection for "+
                                     field);
                           Wrapping.invokeRoleMethod(wrappee,"getCollection",
                                                     new Object[] { field } );
                        }
                        else {
                           // test if collection must be loaded here
                           // instead of being loaded when accessed
                           if (force) {
                              getCollection(wrappee,(CollectionItem)field);
                           } else {
                              Log.trace("persistence","storing OID for collection "+
                                        field+"="+values[i].value);
                              notloadedReferences.put(field,values[i].value);
                           }
                        }
                     }
                  } catch (Exception e) {
                     Log.error("Failed to load field "+field.getLongName()+" for "+oid);
                     e.printStackTrace();
                  }
               }
            }
         }
      }
   }
   
   /**
    * This role method wraps all the collection of the current wrappee
    * so that they will trigger the needed loads from the storage when
    * they are accessed.
    *
    * @param isNew must true if the object is newly created, false if
    * loaded from storage */

   public void wrapCollections(Wrappee wrappee, boolean isNew) {
      Log.trace("persistence","wrapCollection("+wrappee.getClass().getName()+
                "), isNew="+isNew);
      // Wrap collections
      try {
         ClassItem cli = ClassRepository.get().getClass(wrappee);
         Vector collectionsToInit = new Vector();
         CollectionItem[] collections = cli.getCollections();
         for (int i=0; i<collections.length; i++) {
            if (collections[i].isTransient())
               continue;
            Object coll_value = collections[i].get(wrappee);
            Class collType = collections[i].getType();
            if (coll_value == null) {
               Log.warning("persistence.wrap",
                           "uninitialized collection "+collections[i].getName());
               // uninitialized field
               if (collType.isArray()) {
                  coll_value = java.lang.reflect.Array.newInstance(
                     collType.getComponentType(),0);
               } else {
                  try {
                     // This will fail if the declared type of the
                     // attribute is an interface
                     if (!collType.isInterface()) {
                        coll_value = collType.newInstance();
                     }
                  } catch(Exception e) {
                     e.printStackTrace();
                  }
               }
               collections[i].set(wrappee,coll_value);
            }
            Log.trace("persistence.wrap","");
            Class actualCollType = coll_value.getClass();
            if ( //isWrapped(collections[i]) || 
                 actualCollType == org.objectweb.jac.lib.java.util.Vector.class ||
                 actualCollType == org.objectweb.jac.lib.java.util.HashMap.class ||
                 actualCollType == org.objectweb.jac.lib.java.util.Hashtable.class ||
                 actualCollType == org.objectweb.jac.lib.java.util.HashSet.class ) 
            {
               OID coll_oid;
               if (isNew) {
                  coll_oid = getStorage().createObject(collections[i].getType().getName());
               } else {
                  coll_oid = (OID)notloadedReferences.get(collections[i]);
                  if (coll_oid==null)
                     coll_oid = (OID)getStorage().getField(getOID(wrappee),collections[i]);
                  else
                     notloadedReferences.remove(collections[i]);
                  // classes may change after objects were created and
                  // made persistent
                  if (coll_oid==null) {
                     coll_oid = getStorage().createObject(collections[i].getType().getName());
                     initCollection(wrappee,collections[i]);
                  }
               }
               
               Log.trace("persistence",2,collections[i]+" -> "+coll_oid);

               boolean preloaded = PersistenceAC.isFieldPreloaded(collections[i]);
               if (collections[i].isList()) {
                  // lists
                  Log.trace("persistence.wrap",2,
                            "wrapping List "+collections[i].getName()+
                            " "+System.identityHashCode(coll_value));
                  ListWrapper wrapper = new ListWrapper(getAspectComponent(),preloaded);
                  Wrappee list = (Wrappee)coll_value;
                  setOID(list,coll_oid);
                  Wrapping.wrap(list,null,wrapper);
                  String[] accessors = 
                     new String[] {"add","clear","contains",
                                   "get","remove","set","size",
                                   "indexOf","lastIndexOf","isEmpty"};
                  Wrapping.wrapSameName(list,wrapper,accessors);
                  Wrapping.wrap(list,wrapper,"memorizeUseDate",accessors);
                  String[] preloads = new String[] {/*"equals",*/"toArray","iterator"};
                  if (collections[i].getAttribute(PersistenceAC.NO_CACHE)!=null) {
                     Wrapping.wrapSameName(list,wrapper,preloads);
                  } else {
                     Wrapping.wrap(list,wrapper,"preload",preloads);
                     //Wrapping.wrap(list,wrapper,"memorizeUseDate",preloads);
                  }
               } else if (collections[i].isSet()) {
                  // sets
                  Log.trace("persistence.wrap",2,
                            "wrapping Set "+collections[i].getName());
                  SetWrapper wrapper = new SetWrapper(getAspectComponent(),preloaded);
                  Wrappee set = (Wrappee)coll_value;
                  setOID(set,coll_oid);
                  if (set != null) {
                     Wrapping.wrap(set,null,wrapper);
                     String[] accessors=new String[] {"add","clear","contains",
                                                      "remove","size","isEmpty"};
                     Wrapping.wrapSameName(set,wrapper,accessors);
                     //Wrapping.wrap(set,wrapper,"memorizeUseDate",accessors);
                     String[] preloads=new String[] {/*"equals",*/"toArray","iterator"};
                     if (collections[i].getAttribute(PersistenceAC.NO_CACHE)!=null) {
                        Wrapping.wrapSameName(set,wrapper,preloads);
                     } else {
                        Wrapping.wrap(set,wrapper,"preload",preloads);
                        //Wrapping.wrap(set,wrapper,"memorizeUseDate",preloads);
                     }
                  } else {
                     Log.warning("persistence",
                                 "Set "+cli.getName()+"."+collections[i].getName()+
                                 " is null, not wrapped");
                  }
               } else if (collections[i].isMap()) {
                  // maps
                  Log.trace("persistence.wrap",2,
                            "wrapping Map "+collections[i].getName());
                  MapWrapper wrapper = new MapWrapper(getAspectComponent(),preloaded);
                  Wrappee map = (Wrappee)coll_value;
                  setOID(map,coll_oid);
                  if (map != null) {
                     Wrapping.wrap(map,null,wrapper);
                     String[] accessors=
                        new String[] {"clear","isEmpty","size",
                                      "containsKey","containsValue","get","put","remove"};
                     Wrapping.wrapSameName(map,wrapper,accessors);
                     //Wrapping.wrap(map,wrapper,"memorizeUseDate",accessors);
                     String[] preloads=new String[] {/*"equals",*/"keySet","entrySet","values"};
                     Wrapping.wrap(map,wrapper,"preload",preloads);
                     //Wrapping.wrap(map,wrapper,"memorizeUseDate",preloads);
                  } else {
                     Log.warning("persistence",
                                 "Map "+cli.getName()+"."+collections[i].getName()+
                                 " is null, not wrapped");
                  }
               }
            }
         }
      } catch(Exception e) {
         e.printStackTrace();
      }
   }

   /**
    * This wrapping method apply the persistence aspect for a given
    * call on a persistent object.
    *
    * <p>It uses the RTTI aspect to determine the accessed and written
    * field types.<p>
    *
    * Default behavior is:<p>
    *
    * <ul>
    * <li>get the fields that are accessed for reading by this wrapped
    * method
    * <li>if the field is a collection, get the collection from the
    * storage
    * <li>if the field is a reference, get the referenced object from
    * the storage
    * <li>proceed the wrappee (and other wrappers if needed)
    * <li>get the fields that were written by this wrapped method
    * <li>if the field is a collection, update the storage depending
    * on the wrapped method type (adder or remover)
    * <li>if the field is a reference, update the storage to change
    * the referenced object
    * <li> if the field is a simple field, just update the storage
    * with its value.
    * </ul>
    *
    * @return the value returned by the wrapped method
    *
    * @see org.objectweb.jac.core.rtti.MethodItem#getAccessedFields()
    * @see #getCollection(Wrappee,CollectionItem)
    * @see #getReference(Wrappee,FieldItem)
    * @see #addToCollection(Interaction,CollectionItem,Object)
    * @see #removeFromCollection(Interaction,CollectionItem,Object)
    * @see #setReference(Wrappee,FieldItem,Wrappee)
    * @see #setField(Wrappee,FieldItem)
    * @see org.objectweb.jac.core.rtti.MethodItem#getWrittenFields()
    * @see org.objectweb.jac.core.rtti */

   public Object applyPersistence(Interaction interaction) throws Exception 
   {
      if (interaction.method.isStatic()) {
         return proceed(interaction);
      }
      Log.trace("persistence.apply",
                "applyPersistence on "+getOID(interaction.wrappee)+
                "("+interaction.wrappee+")."+interaction.method);
      MethodItem method = (MethodItem)interaction.method;
      FieldItem[] fields;

      // the object may be wrapped but not persistent yet
      if (getOID(interaction.wrappee)!=null ) {

         ClassItem cli = ClassRepository.get().getClass(interaction.wrappee.getClass());
         // preload accessed fields and collections
         fields = method.getAccessedFields();
         for( int i = 0; fields != null && i < fields.length; i++ ) {
            FieldItem field = fields[i];
            
            if (!field.isTransient() && !field.isFinal()) {
               if (field instanceof CollectionItem && 
                   !isWrapped(interaction.wrappee,field)) {
                  getCollection(interaction.wrappee,(CollectionItem)field);
               } else if (field.isReference()) {
                  getReference(interaction.wrappee,field);
               } 
            }
         }  
      }

      Object result = proceed(interaction);

      if (getOID(interaction.wrappee)!=null ) {
      
         // save written fields
         fields = method.getWrittenFields();
         for( int i = 0; fields != null && i < fields.length; i++ ) {
            FieldItem field = fields[i];
            if (!field.isTransient()) {
               Log.trace("persistence",
                         "Save written field in "+interaction.method+" : "+field.getName());
               if (field.isReference()) {
                  setReference(interaction.wrappee,field,
                               (Wrappee)field.get(interaction.wrappee));
               } else if (field instanceof CollectionItem) {
                  // Nothing to do if the collection is wrapped
                  // BUT THE COLLECTION MUST BE SAVED IN THE CASE OF ARRAYS
                  //     storage.updateCollection(collection);
                  CollectionItem collection = (CollectionItem)field;
                  if (collection.isArray() &&
                      ((MethodItem)interaction.method).getCollectionIndexArgument()!=-1) {
                     updateCollection(
                        collection,
                        ((MethodItem)interaction.method).getCollectionIndexArgument());
                  }
               } else {
                  setField(interaction.wrappee,field);
               }
            }
         }
         
         // handle added collections (adder)
         CollectionItem[] collections = method.getAddedCollections();
         for( int i = 0; collections != null && i < collections.length; i++ ) {
            if ( !collections[i].isTransient() && 
                 !isWrapped(interaction.wrappee,collections[i]) ) {
               addToCollection(interaction,collections[i],interaction.args[0]);
            }
         }
         // handle removed collections (remover)
         collections = method.getRemovedCollections();
         for( int i = 0; collections != null && i < collections.length; i++ ) {
            if ( !collections[i].isTransient() && 
                 !isWrapped(interaction.wrappee,collections[i]) ) {
               removeFromCollection(interaction,collections[i],interaction.args[0]);
            }
         }
         
         // UNCOMMENT THIS WHEN UNWRAP WILL BE THREAD-SAFE!!!
         // unwrap the method if it was only a getter
         //if ( ! method.hasAddedCollections() &&
         //     ! method.hasRemovedCollections() &&
         //     ! method.hasWrittenFields() )
         //{
         //   Log.trace("persistence.ref","unwrapping "+getOID()+" "+method);
         //   Wrapping.unwrap(interaction.wrappee,this,"applyPersistence",method);
         //}
      }
      return result;
   }
   
   /**
    * This method loads alls the persistent objects that are contained
    * in this collection from the storage.
    *
    * <p>If the collection has already been loaded from the storage,
    * do nothing.<p>
    *
    * @param collection the collection to load */

   public void getCollection(Wrappee wrappee, CollectionItem collection) 
      throws Exception
   {
      checkOid(wrappee);
      Log.trace("persistence",
                getOID(wrappee)+".getCollection("+collection.getName()+")");
      if (!loadedVectors.contains(collection.getName())) {
         String classID = collection.getParent().getName();
         List vector=null;
         Map map=null;
         OID cid = getStorage().getCollectionID(getOID(wrappee),collection);
         Log.trace("persistence.collection","cid="+cid);
         if (cid==null) {
            // Handle the case of a new collection added in the model
            initCollection(wrappee,collection);
         }
         if (collection.isList() || collection.isArray()) {
            vector = getStorage().getList(getOID(wrappee),collection);
         } else if (collection.isSet()) {
            vector = getStorage().getSet(getOID(wrappee),collection);
         } else if (collection.isMap()) {
            map = getStorage().getMap(getOID(wrappee),collection);
         } else {
            Log.error("unhandled collection type : "+collection.getType());
            Log.error("please see the JAC programming guidelines for more details");
            //new Exception().printStackTrace();
            return;
         }

         collection.clear(wrappee);

         if (collection.isMap()) {
            Iterator it = map.entrySet().iterator();
            while (it.hasNext()) {
               Map.Entry entry = (Map.Entry)it.next();
               attrdef(ATTR_ADDED,"true");
               collection.add(wrappee, 
                              normalizeOutput(entry.getValue()), 
                              normalizeOutput(entry.getKey()));
               attrdef(ATTR_ADDED,null);
            }
         } else {
            for (int i=0; i<vector.size(); i++) {
               attrdef(ATTR_ADDED,"true");
               Log.trace("persistence.collection","adding "+
                         normalizeOutput(vector.get(i))+" to "+
                         collection+"(wrappee="+wrappee+")");
               collection.add(wrappee, normalizeOutput(vector.get(i)), null);
               attrdef(ATTR_ADDED,null);
            }
         }
         loadedVectors.add(collection.getName());
      }
   }

   /**
    * This method stores the persistent object that is added to the
    * collection into the storage.<p>
    *
    * @param collection the collection to add to
    *
    * @see #removeFromCollection(Interaction,CollectionItem,Object)
    * @see #getCollection(Wrappee,CollectionItem)
    */   

   protected void addToCollection(Interaction interaction, 
                                  CollectionItem collection, Object value ) 
      throws Exception 
   {
      checkOid(interaction.wrappee);
      Log.trace("persistence",
                getOID(interaction.wrappee)+".addToCollection: "+
                interaction.method+Arrays.asList(interaction.args));

      if ( isPersistent(interaction.wrappee) && (interaction.args.length == 1) )
      {
         // the method's arguments look correct
         value = normalizeInput(value);
         OID cid = getStorage().getCollectionID(getOID(interaction.wrappee), collection);
         if ( collection.isList() || collection.isArray() ) {
            getStorage().addToList( cid, value );
         } else if ( collection.isSet() ) {
            getStorage().addToSet( cid, value );
         } else {
            Log.error("unhandled collection type : "+collection.getType());
            new Exception().printStackTrace();
         }
         
      } else {
         Log.error("NOT IMPLEMENTED YET !!!");
         // the method's arguments are weird, so let's diff
         //         storage.updateCollection(collection);
      }
   }

   /**
    * This method delete the persistent object that is removed from
    * the collection from the storage.<p>
    *
    * @param collection the collection to remove from
    *
    * @see #addToCollection(Interaction,CollectionItem,Object)
    * @see #getCollection(Wrappee,CollectionItem) */

   protected void removeFromCollection(Interaction interaction, 
                                       CollectionItem collection, Object value ) 
      throws Exception 
   {
      checkOid(interaction.wrappee);
      if (isPersistent(interaction.wrappee)) {

         Log.trace("persistence",getOID(interaction.wrappee)+
                   ".removeFromCollection "+interaction.method);

         value = normalizeInput(value);
         OID cid = getStorage().getCollectionID(getOID(interaction.wrappee),
                                                collection);
         if ( collection.isList() || collection.isArray() ) {
            getStorage().removeFromList( cid, value );
         } else if ( collection.isSet() ) {
            getStorage().removeFromSet( cid, value );
         } else {
            Log.error("unhandled collection type : "+collection.getType());
            new Exception().printStackTrace();
         }
      }
   }

   /**
    * Update the element of a collection at a gievn position
    */
   public void updateCollection(CollectionItem collection, int position) 
      throws Exception 
   {
      /*
      checkOid();
      Log.trace("persistence",
                getOID()+".updateCollection: "+method()+Arrays.asList(args()));

      if (isPersistent())
      {
         // the method's arguments look correct
         value = normalizeInput(value);
         OID cid = getStorage().getCollectionID(getOID(), collection);
         if ( collection.isList() || collection.isArray() ) {
            Object value = collection.get(wrappee(),position);
            getStorage().setListItem(cid, position, value);
         } else {
            Log.error("unhandled collection type : "+collection.getType());
         }
      }     
      */
   }

   /**
    * This method loads the persistent object that is pointed by the
    * reference from the storage.<p>
    *
    * If the reference has already been loaded from the storage, do
    * nothing.<p>
    *
    * @param reference the reference to load
    *
    * @see #setReference(Wrappee,FieldItem,Wrappee) */

   public void getReference(Wrappee wrappee, FieldItem reference) 
      throws Exception 
   {
      try {
         checkOid(wrappee);
         Log.trace("persistence.ref",
                   getOID(wrappee)+"["+this+"].getReference("+
                   reference.getName()+")");
         if (isPersistent(wrappee)) {
            if (!loadedReferences.contains(reference.getName())) {
               OID lOid = (OID)notloadedReferences.get(reference);
               if (lOid==null)
                  lOid = (OID)getStorage().getField(getOID(wrappee),reference);
               else
                  notloadedReferences.remove(reference);
               Log.trace("persistence.ref",
                         getOID(wrappee)+"."+reference.getName()+
                         " -> "+lOid+" ("+wrappee.getClass().getName()+")");
               Object obj = null;
               if (lOid != null) {
                  // What the hell is this ???  It bugs if a reference
                  // is initialized by the default constructor
                  //obj=getAC().getObject(lOid,reference.get(wrappee));
                  obj = getAC().getObject(lOid,null);
               }
               attrdef(ATTR_ADDED,"true");
               try {
                  reference.set(wrappee,obj);
               } catch (IllegalArgumentException illegal) {
                  // This may happen if the type of the reference has been changed
                  Log.error("Incompatible types "+reference.getType()+" and "+obj);
               } finally {
                  attrdef(ATTR_ADDED,null);
               }
               loadedReferences.add(reference.getName());
            } else {
               Log.trace("persistence.ref",2,
                         getOID(wrappee)+"."+reference.getName()+" already loaded");
            }
         } else {
            Log.trace("persistence.ref",
                      getOID(wrappee)+" is not persistent, "+
                      reference.getName()+" not loaded");
         }
      } catch (Exception e) {
         Log.error("getReference");
         e.printStackTrace();
      }
   }

   /**
    * This method stores the persistent object that is pointed by the
    * reference into the storage.<p>
    *
    * @param reference the reference to store
    *
    * @see #getReference(Wrappee,FieldItem) */

   protected void setReference(Wrappee wrappee, FieldItem reference, Wrappee value ) 
      throws Exception { 
      
      if( value != null ) {
         checkOid(wrappee);
         if (isPersistent(wrappee)) {
            Wrapping.invokeRoleMethod(value,"makePersistent",new Object[0]);
            if ( getOID(wrappee) == null ) {
               // this should never happen
               //               Log.error("oid is NULL in setReference, method = "+interaction.method);
               throw new Exception("Null OID Exception");
            }
            getStorage().updateField(getOID(wrappee),reference,getOID(value));
         }
      } else {
         getStorage().updateField(getOID(wrappee),reference,null);
      }
   }

   /**
    * This method stores the field into the storage.<p>
    *
    * @param field the field to store */

   protected void setField(Wrappee wrappee, FieldItem field) throws Exception { 
      checkOid(wrappee);
      if ( isPersistent(wrappee) ) {
         getStorage().updateField( getOID(wrappee), field, field.get(wrappee) ) ;
      }
   }

   public void initCollections(Wrappee wrappee, CollectionItem[] collections) 
      throws Exception 
   {
      Log.trace("persistence", "wrappee = "+wrappee.getClass().getName() );
      Log.trace("persistence", "collections = "+Arrays.asList(collections) );
      //Log.trace("persistence", "collections.length = "+collections.length );
      for ( int i = 0; i < collections.length; i++ ) {
         if (!collections[i].isTransient()) {
            Log.trace("persistence", "collections["+i+"] = "+collections[i]);
            initCollection(wrappee,collections[i]);
         }
      }
   }

   void initCollection(Wrappee wrappee, CollectionItem collection) throws Exception {
      Log.trace("persistence", "Init collection "+collection.getName());
      loadedVectors.add(collection.getName());
      Storage storage = getStorage();
      if ( collection.isList() || 
           collection.isSet() || 
           collection.isArray() ) {
         OID cid;
         if (isWrapped(wrappee,collection)) {
            cid = getOID((Wrappee)collection.get(wrappee));
         } else {
            cid = storage.createObject(collection.getType().getName());
         }
         storage.setField(getOID(wrappee),collection,cid);
         Collection coll = collection.getActualCollection(wrappee);
         Log.trace("persistence","Coll value = " + coll );
         if (coll==null) {
            Log.warning("uninitialized collection "+collection+" detected by persistence");
            return;
         }
         Iterator it = coll.iterator();
         while (it.hasNext()) {
            Object value = normalizeInput(it.next());
            Log.trace("persistence","Coll elt value = " + value );
            if (value!=null) {
               if ( collection.isList() || collection.isArray() ) {
                  storage.addToList(cid, value);
               } else {
                  storage.addToSet(cid, value);
               }
            }
         }
      } else if (collection.isMap()) {
         OID cid;
         if (isWrapped(wrappee,collection)) {
            cid = getOID((Wrappee)collection.get(wrappee));
         } else {
            cid = storage.createObject(collection.getType().getName());
         }
         storage.setField(getOID(wrappee),collection,cid);
         Map map = (Map)collection.get(wrappee);
         Set entries = map.entrySet();
         Iterator it = entries.iterator();
         while ( it.hasNext() ) {
            Map.Entry entry = (Map.Entry)it.next();
            storage.putInMap(cid,normalizeInput(entry.getKey()),
                                  normalizeInput(entry.getValue()));
         }
      } else {
         Log.error("unhandled collection type : "+collection.getType());
         new Exception().printStackTrace();
      }
   }

   /**
    * Initialize all fields of an object in the storage.
    *
    * <p>This method is called when a new object is made persistent. The
    * references and collections are also recursively saved into the
    * storage.<p>
    *
    * @param oid the object ID in the storage
    * @param wrappee the initializer object
    */

   public void initAllFields(Wrappee wrappee) throws Exception {
      checkOid(wrappee);
      String classID = wrappee.getClass().getName();
      Log.trace("persistence","initAllfields("+getOID(wrappee)+") : "+classID);
      ClassItem cli = ClassRepository.get().getClass(classID);
      FieldItem[] fields = cli.getPrimitiveFields();
      Storage storage = getStorage();
      storage.startTransaction();
      try {
         // fields
         for ( int i = 0; i < fields.length; i++ ) {
            if (!fields[i].isTransient() 
                && !fields[i].isStatic() 
                && !fields[i].isFinal()) 
            {
               Object fieldValue = fields[i].get(wrappee);
               Log.trace("persistence",
                         "Init field "+fields[i].getName()+"->"+fieldValue);
               if ( fieldValue!=null ) {
                  storage.setField(getOID(wrappee),fields[i],fieldValue);
               }
            }
         } 

         FieldItem[] references = cli.getReferences();
         
         // references
         for (int i=0; i<references.length; i++) {
            if (!references[i].isTransient()
                && !references[i].isStatic() 
                && !references[i].isFinal()) 
            {
               Log.trace("persistence","Init reference " + references[i].getName() );
               Wrappee obj = (Wrappee) references[i].getActualField().get(wrappee);
               Log.trace("persistence","Ref value = " + 
                         ( obj!=null ? obj.getClass().getName() : "null" ) );
               
               loadedReferences.add(references[i].getName());
               if (obj!=null) {
                  Wrapping.invokeRoleMethod(obj,"makePersistent",new Object[0]);
                  if (getOID(obj) == null) {
                     // this should never happen
                     Log.error("wrapper.oid is NULL in initAllFields, reference = "+
                               references[i].getName());
                  } else {
                     Log.trace("persistence","Ref value = " + getOID(obj));
                     storage.setField(getOID(wrappee),references[i],getOID(obj));
                     Log.trace("persistence","Ref saved" );
                  }
               }
            }
            
            // UNCOMMENT THIS WHEN UNWRAP WILL BE THREAD-SAFE!!!
            // unwrap the reference's getter because it won't need to be
            // reloaded from the storage
            //MethodItem getter = references[i].getGetter();
            //if (getter!=null) {
            //   Wrapping.unwrap(interaction.wrappee,this,"applyPersistence",getter);            
            //}
         }

         initCollections(wrappee,cli.getCollections());
      } catch(Exception e) {
         storage.rollback();
      }
      storage.commit();
   }

   public Storage getStorage() {
      try {
         return getAC().getStorage();
      } catch (Exception e) {
         Log.error("getStorage failed: "+e);
         e.printStackTrace();
         return null;
      }
   }

   public final PersistenceAC getAC() {
      return (PersistenceAC)getAspectComponent();
   }

   public void checkOid(Wrappee wrappee) throws InvalidOidException
   {
      if (getOID(wrappee)==null) 
         throw new InvalidOidException("oid is NULL");
   }

   public static class InvalidOidException extends RuntimeException {
      public InvalidOidException(String msg) { super(msg); }
   }
}

// Local Variables: ***
// c-basic-offset:3 ***
// End: ***
