/*
  Copyright (C) 2001-2003 Laurent Martelli <laurent@aopsys.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package org.objectweb.jac.aspects.persistence;

import java.sql.*;
import java.util.*;
import java.lang.reflect.*;
import org.objectweb.jac.core.rtti.*;
import org.objectweb.jac.util.*;

/**
 * Implements the storage to store within an SQL compliant database system.
 *
 * @see LongOID
 */

public abstract class SQLStorage implements Storage,java.io.Serializable {

   /**
    * The SQL connection to the database that is use by this storage. */
   protected Connection db;

   /**
    * Default constructor. */
   protected SQLStorage() throws SQLException {}

   /**
    * Creates a new SQL storage.<p>
    *
    * @param db the connection to the database
    */

   public SQLStorage(Connection db) throws SQLException {
      setConnection(db);
   }

   public void close() {}

   /**
    * Resets the connection to the database with the given
    * connection.<p>
    * 
    * @param db the new connection */

   protected void setConnection(Connection db) {
      try {
         this.db = db;
      } catch (Exception e) {
         Log.error("Failed to create SQL statement : "+e);
      }
   }

   protected int executeUpdate(String query) throws SQLException {
      return db.createStatement().executeUpdate(query);
   }

   protected ResultSet executeQuery(String query) throws SQLException {
      return db.createStatement().executeQuery(query);
   }

   protected boolean execute(String query) throws SQLException {
      return db.createStatement().execute(query);
   }

   public void deleteObject(OID oid) throws Exception
   {
      Log.trace("storage","deleteObject("+oid+")");
      executeUpdate("delete from objects where id="+oid);
      executeUpdate("delete from roots where id='"+oid+"'");
      // WE SHOULD ALSO REMOVE COLLECTIONS
   }

   public void setField(OID oid, FieldItem field, Object object) 
      throws Exception
   {
      Log.trace("storage","setField("+oid+","+field.getName()+","+object+")");
      String value = ValueConverter.objectToString(object);
      String fieldID = field.getName();
      String query = "insert into objects (id,fieldID,value) values "+
         "("+oid+",'"+fieldID+"','"+addSlashes(value)+"')";
      if (executeUpdate(query)==0) {
         Log.error("setField failed : "+oid+","+fieldID+","+value);
      }
   }

   public void updateField(OID oid, FieldItem field, Object object) 
      throws Exception
   {
      Log.trace("storage","updateField("+oid+","+field+","+object+")");
      String fieldID = field.getName();
      String value = ValueConverter.objectToString(object);
      String query = "update objects set value='"+addSlashes(value)+"' where "+
         "id="+oid+" and fieldID='"+fieldID+"'";
      if (executeUpdate(query)==0) {
         setField(oid,field,object);
      }
   }

   public Object getField(OID oid, FieldItem field) 
      throws Exception
   {
      Log.trace("storage","getField("+oid+","+field.getName()+")");
      checkStorage();
      String fieldID = field.getName();
      ResultSet rs = executeQuery("select value from objects where id="+oid+
                                  " and fieldID='"+fieldID+"'");
      if (rs.next()) {
	 return ValueConverter.stringToObject(rs.getString("value"));
      } else {
         if (field.isPrimitive()) {
            Log.warning("no such field in storage "+oid+","+fieldID);
         }
	 return null;
      }
   }

   public StorageField[] getFields(OID oid, ClassItem cl, FieldItem[] fields) 
      throws Exception
   {
      Log.trace("storage","getFields "+oid+","+cl+","+Arrays.asList(fields));
      // compute stringified list of fields for SQL query
      if (fields.length == 0) {
         return new StorageField[0];
      }
      String fieldlist = "(";
      for (int i=0; i<fields.length; i++) {
         if (i!=0)
            fieldlist += " or ";
         fieldlist += "fieldID='"+fields[i].getName()+"'";
      }
      fieldlist += ")";
      
      StorageField fieldValues[] = new StorageField[fields.length];
      String query = "select * from objects where id="+oid+" and "+fieldlist;
      Log.trace("sql",query);
      ResultSet rs = executeQuery(query);

      int i=0;
      while (rs.next()) {
         FieldItem field = cl.getField(rs.getString("fieldID"));
         fieldValues[i] = new StorageField(
            cl,field,
            ValueConverter.stringToObject(rs.getString("value")));
         i++;
      }
      return fieldValues;
   }

   public void removeField(OID oid, FieldItem field, Object value) 
      throws Exception
   {
      Log.trace("storage","removeField("+oid+","+field+","+value+")");
      String fieldID = field.getName();
      executeUpdate("delete from objects where id="+oid+
		       " and fieldID='"+fieldID+"'");      
   }

   public Collection getRootObjects() throws Exception {
      Log.trace("storage","getRootObjects");
      String sql = "select id from roots";
      ResultSet rs = executeQuery(sql);
      Vector result = new Vector();
      while (rs.next()) {
         result.add(new LongOID(rs.getLong("value")));
      }
      Log.trace("storage","getRootObjects returns " + result);
      return result;
   }

   // Collection methods

   public OID getCollectionID(OID oid, CollectionItem collection) 
      throws Exception 
   {
      return getOID("select value from objects where "+
                    "id="+oid+" and fieldID='"+collection.getName()+"'");
   }

   public List getCollectionValues(OID oid, CollectionItem collection, 
                                   String table, String orderBy) 
      throws Exception
   {
      Log.trace("storage","getCollectionValues("+oid+","+collection+")");
      String fieldID = collection.getName();

      String sql = "select "+table+".value from "+table+",objects where "+
         "objects.id="+oid+
	 " and objects.fieldID='"+fieldID+"'"+
         " and objects.value="+table+".id";
      if ( orderBy!=null) {
         sql += " order by " + orderBy;
      }
      
      ResultSet rs = executeQuery(sql);
      Vector result = new Vector();
      while (rs.next()) {
         result.add(ValueConverter.stringToObject(rs.getString("value")));
      }
      Log.trace("storage","getCollectionValues returns " + result);
      return result;
   }

   public boolean collectionContains(String table, OID cid, Object value) 
      throws Exception 
   {
      ResultSet res = executeQuery(
         "select id from "+table+" where "+
         "id="+cid+" and value='"+
            addSlashes(ValueConverter.objectToString(value))+"'");
      return res.next();
   }

   // List methods

   public void clearList(OID cid) 
      throws Exception 
   {
      Log.trace("storage","clearList("+cid+")");
      executeUpdate("delete from lists where id="+cid);
   }

   public List getList(OID oid, CollectionItem collection) 
      throws Exception
   {
      return getList(getCollectionID(oid,collection));
   }

   public List getList(OID cid)
      throws Exception
   {
      Log.trace("storage","getList("+cid+")");
      ResultSet rs = executeQuery("select value from lists "+
                                     "where id="+cid+" order by index");
      Vector result = new Vector();
      while (rs.next()) {
         result.add(ValueConverter.stringToObject(rs.getString("value")));
      }
      Log.trace("storage","getList returns " + result);
      return result;
   }

   public long getListSize(OID cid) 
      throws Exception
   {
      return getLong("select count(*) from lists where id="+cid);
   }

   public boolean listContains(OID cid, Object value) 
      throws Exception 
   {
      return collectionContains("lists",cid,value);
   }

   public Object getListItem(OID cid, long index)
      throws Exception
   {
      ResultSet rs = executeQuery("select value from lists where "+
                                  "id="+cid+" order by index limit 1,"+index);
      if (rs.next()) {
         return ValueConverter.stringToObject(rs.getString("value"));
      } else {
         return null;      
      }
   }

   public long getIndexInList(OID cid, Object value)
      throws Exception
   {
      ResultSet rs = executeQuery(
         "select min(index) as index from lists where "+
         "id="+cid+" and value='"+
         addSlashes(ValueConverter.objectToString(value))+"'");
      if (rs.next()) {
         long index = rs.getLong(1);
         return getLong("select count(*) from lists where id="+cid+
                        " and index<="+index)-1;
      } else {
         return -1;
      }
   }

   protected long getInternalIndexInList(OID cid, Object value)
      throws Exception
   {
      ResultSet rs = executeQuery(
         "select min(index) as index from lists where "+
         "id="+cid+" and value='"+
         addSlashes(ValueConverter.objectToString(value))+"'");
      if (rs.next()) {
         long result = rs.getLong(1);
         if (rs.wasNull())
            return -1;
         else
            return result;
      } else {
         return -1;
      }
   }

   public long getLastIndexInList(OID cid, Object value)
      throws Exception
   {
      ResultSet rs = executeQuery(
         "select max(index) from lists where "+
         "id="+cid+" and value='"+
         addSlashes(ValueConverter.objectToString(value))+"'");
      if (rs.next()) {
         long index = rs.getLong(1);
         return getLong("select count(*) from lists where id="+cid+
                        " and index<="+index)-1;
      } else {
         return -1;      
      }
   }

   public void addToList(OID cid, long position, Object value)
      throws Exception
   {
      Log.trace("storage","addToList("+cid+","+position+","+value+")");
      executeUpdate("update lists set index=index+1 where id="+cid+
                    " and index>="+position);
      executeUpdate("insert into lists (id,index,value) values "+
                    "("+cid+","+position+",'"+
                       addSlashes(ValueConverter.objectToString(value))+"')");
   }

   public void addToList(OID cid, Object value)
      throws Exception
   {
      Log.trace("storage","addToList("+cid+","+value+")");
      long size = getListSize(cid);
      String indexExpr;
      if (size==0)
         indexExpr = "0";
      else
         indexExpr = "select max(index)+1 from lists where id="+cid;
      executeUpdate("insert into lists (id,index,value) values "+
                    "("+cid+",("+indexExpr+"),'"+
                       addSlashes(ValueConverter.objectToString(value))+"')");
   }

   public void setListItem(OID cid, long index, Object value)
      throws Exception
   {
      Log.trace("storage","setListItem("+cid+","+index+","+value+")");
      executeUpdate("update lists set value='"+
                       addSlashes(ValueConverter.objectToString(value))+
                    " where id="+cid+" and index="+index);
   }

   public void removeFromList(OID cid, long position)
      throws Exception
   {
      Log.trace("storage","removeFromList("+cid+","+position+")");
      // First, get the index for the position
      ResultSet rs = executeQuery("select index from lists where "+
                                  "id="+cid+" order by index limit 1,"+position);
      long index = rs.getLong("index");
      executeUpdate("delete from lists where "+"id="+cid+" and index="+index);
   }

   public void removeFromList(OID cid, Object value)
      throws Exception
   {
      Log.trace("storage","removeFromList("+cid+","+value+")");
      long index = getInternalIndexInList(cid,value);
      executeUpdate("delete from lists where "+"id="+cid+" and index="+index);
   }

   // Set methods

   public void clearSet(OID cid) 
      throws Exception 
   {
      Log.trace("storage","clearSet("+cid+")");
      executeUpdate("delete from sets where id="+cid);
   }

   public List getSet(OID oid, CollectionItem collection) throws Exception {
      return getSet(getCollectionID(oid,collection));
   }   

   public List getSet(OID cid) throws Exception {
      Log.trace("storage","getSet("+cid+")");
      ResultSet rs = executeQuery("select value from sets "+
                                     "where id="+cid);
      Vector result = new Vector();
      while (rs.next()) {
         result.add(ValueConverter.stringToObject(rs.getString("value")));
      }
      Log.trace("storage","getSet returns " + result);
      return result;
   }   

   public long getSetSize(OID cid)
      throws Exception
   {
      return getLong("select count(*) from sets where id="+cid);
   }

   public boolean setContains(OID cid, Object value) 
      throws Exception 
   {
      return collectionContains("sets",cid,value);
   }

   public boolean addToSet(OID cid, Object value) 
      throws Exception 
   {
      Log.trace("storage","addToSet("+cid+","+value+")");
      if (!collectionContains("sets",cid,value)) {
         executeUpdate("insert into sets (id,value) values "+
                       "("+cid+",'"+
                       addSlashes(ValueConverter.objectToString(value))+"')");
         return true;
      } else {
         return false;
      }
   }

   public boolean removeFromSet(OID cid, Object value) 
      throws Exception 
   {
      Log.trace("storage","removeFromSet("+cid+","+value+")");
      boolean result = collectionContains("set",cid,value);
      if (result)
         executeUpdate(
            "delete from sets where "+
            "id="+cid+" and value='"+
            addSlashes(ValueConverter.objectToString(value))+"'");
      return result;
   }

   // Map functions

   public void clearMap(OID cid) 
      throws Exception 
   {
      Log.trace("storage","clearMap("+cid+")");
      executeUpdate("delete from maps where id="+cid);
   }

   public Map getMap(OID oid, CollectionItem collection) 
      throws Exception
   {
      return getMap(getCollectionID(oid,collection));
   }

   public Map getMap(OID cid) throws Exception
   {
      Log.trace("storage","getMap("+cid+")");
      ResultSet rs = executeQuery("select value,key from maps where id="+cid);
      Map result = new HashMap();
      while (rs.next()) {
         result.put(ValueConverter.stringToObject(rs.getString("key")),
                    ValueConverter.stringToObject(rs.getString("value")));
      }
      Log.trace("storage","getMap returns " + result);
      return result;
   }

   public long getMapSize(OID cid)
      throws Exception
   {
      return getLong("select count(*) from maps where id="+cid);
   }

   public Object putInMap(OID cid, Object key, Object value) 
      throws Exception
   {
      Log.trace("storage","putInMap("+cid+","+key+"->"+value+")");
      if (mapContainsKey(cid,key)) {
         Object old = getFromMap(cid,key);
         executeUpdate(
            "update maps set "+
            "key='"+addSlashes(ValueConverter.objectToString(key))+"',"+
            "value='"+addSlashes(ValueConverter.objectToString(value))+"' "+
            "where id="+cid+
            " and key='"+addSlashes(ValueConverter.objectToString(key))+"'");
         return old;
      } else {
         executeUpdate(
            "insert into maps (id,key,value) values "+
            "("+cid+",'"+
            addSlashes(ValueConverter.objectToString(key))+
            "','"+addSlashes(ValueConverter.objectToString(value))+"')");
         return null;
      }
   }

   public Object getFromMap(OID cid, Object key) 
      throws Exception
   {
      Log.trace("storage","getFromMap("+cid+","+key+")");
      ResultSet res = executeQuery(
         "select value from maps where "+
         "id="+cid+" and key='"+
         addSlashes(ValueConverter.objectToString(key))+"'");
      if ( res.next()) {
         return ValueConverter.stringToObject(res.getString("value"));
      } else {
         return null;
      }
   }

   public boolean mapContainsKey(OID cid, Object key) 
      throws Exception
   {
      Log.trace("storage","mapContainsKey("+cid+","+key+")");
      ResultSet res = executeQuery(
         "select value from maps where "+
         "id="+cid+" and key='"+
         addSlashes(ValueConverter.objectToString(key))+"'");
      return res.next();
   }

   public boolean mapContainsValue(OID cid, Object value) 
      throws Exception 
   {
      return collectionContains("maps",cid,value);
   }

   public Object removeFromMap(OID cid, Object key)
      throws Exception

   {
      Log.trace("storage","removeFromMap("+cid+","+key+")");
      if (!mapContainsKey(cid,key)) {
         return null;
      } else {
         Object result = getFromMap(cid,key);
         executeUpdate("delete from maps where "+
                       "id="+cid+" and key='"+
                       addSlashes(ValueConverter.objectToString(key))+"'");
         return result;
      }
   }

   public String newName(String className) throws Exception
   {
      try {
         return "_"+className+getNextVal("name_counter");
      } catch (Exception e) {
         Log.error("newName failed : "+e);
         e.printStackTrace();
         return null;
      }
   }

   public OID getOIDFromName(String name) 
      throws Exception
   {
      ResultSet rs = executeQuery(
         "select id from roots where name='"+name+"'");
      if (rs.next()) {
	 return new LongOID(rs.getLong("id"));
      } else {
	 return null;
      }
   }
   
   public String getNameFromOID(OID oid) throws Exception
   {
      ResultSet rs = executeQuery(
         "select name from roots where id="+oid);
      if (rs.next()) {
	 return rs.getString("name");
      } else {
	 return null;
      }
   }

   public void bindOIDToName(OID oid, String name) throws Exception
   {
      Log.trace("storage","bindOIDToName "+oid+" -> "+name);
      executeUpdate("insert into roots (id,name) values ("+
		       oid+",'"+name+"')");
   }

   public void deleteName(String name) throws Exception
   {
      Log.trace("storage","deleteName("+name+")");
      executeUpdate("delete from roots where name='"+name+"'");
   }

   public String getClassID(OID oid) throws Exception
   {
      ResultSet rs = executeQuery("select classid from classes where id="+oid);
      if (rs.next()) {
         String classID = rs.getString("classid");
         Log.trace("SQLStorage","getClassID("+oid+") -> "+classID);
         return classID;
      } else {
         throw new NoSuchOIDError(oid);
      }
   }

   public Collection getObjects(ClassItem cl) throws Exception
   {
      Log.trace("storage","getObjects("+cl.getName()+")");
      Vector result = new Vector();
      getObjects(cl,result);
      return result;
   }

   protected void getObjects(ClassItem cl, Vector objects) throws SQLException {
      String query = "select id from classes";
      if (cl != null) {
         query += " where classes.classid='"+cl.getName()+"'";
      }
               
      ResultSet rs = executeQuery(query);
      while (rs.next()) {
         objects.add(new LongOID(rs.getLong("id")));
      }

      Iterator i = cl.getChildren().iterator();
      while(i.hasNext()) {
         ClassItem subclass = (ClassItem)i.next();
         getObjects(subclass,objects);
      }
   }

   public void startTransaction() throws SQLException {
      execute("BEGIN TRANSACTION");
   }

   public void commit() throws SQLException {
      execute("COMMIT");
   }

   public void rollback() throws SQLException {
      execute("ROLLBACK");
   }

   /**
    * Throw an exception if storage is null or invalid
    *
    */
   protected void checkStorage() {
      if (db==null) {
	 Log.error("connection is NULL");
         throw new InvalidStorageException("connection is NULL");
      }
   }

   /**
    * Creates a new object using a PostgreSQL sequance.<p>
    *
    * @return the new object OID
    */

   public OID createObject(String className) throws Exception {
      LongOID res = new LongOID(getNextVal("object_id"));
      executeUpdate("insert into classes (id,classid) values ("+
                    res+",'"+className+"')");
      return res;
   }

   /**
    * Returns the next value of a sequence
    */
   public abstract long getNextVal(String sequence) throws Exception;

   public long getLong(String query) throws Exception {
      ResultSet rs = executeQuery(query);
      rs.next();
      return rs.getLong(1);
   }

   public OID getOID(String query) throws Exception {
      return new LongOID(getLong(query));
   }

   public static class InvalidStorageException extends RuntimeException {
      public InvalidStorageException(String msg) { super(msg); }
   }

   /* add '\' before ''' and '\' */
   public static String addSlashes(String str) {
      StringBuffer res = new StringBuffer(str.length());
      for (int i=0; i<str.length();i++) {
         if (str.charAt(i)=='\'' || str.charAt(i)=='\\') {
            res.append('\\');
         }
         res.append(str.charAt(i));
      }
      return res.toString();
   }
}
