/*
  Copyright (C) 2001 Lionel Seinturier

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser Generaly Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package org.objectweb.jac.core.dist.rmi;


import org.objectweb.jac.core.dist.Distd;
import org.objectweb.jac.core.dist.RemoteContainer;
import java.io.IOException;
import java.lang.Runtime;
import java.net.InetAddress;
import java.rmi.Naming;
import java.rmi.RMISecurityManager;
import org.objectweb.jac.util.Log;
import java.io.File;


/**
 * RMIDistd is a jac daemon that supports the RMI communication protocol.
 *
 * Daemons hold containers (only one for the moment) which themselves hold
 * remote objects.
 *
 * @author <a href="http://www-src.lip6.fr/homepages/Lionel.Seinturier/index-eng.html">Lionel Seinturier</a>
 */
 
public class RMIDistd extends Distd {

   /**
    * This method initializes the RMI communication protocol.
    */
   
   public void init() {

      /** Create and install a security manager */
       
      if (System.getSecurityManager() == null)
         System.setSecurityManager(new RMISecurityManager());

      try {
         File jar = new File(org.objectweb.jac.core.Jac.getJacRoot()+"org.objectweb.jac.jar");
         Log.trace("dist.rmi",1,"$JAC_ROOT/org.objectweb.jac.jar = "+jar.getPath());
         if (jar.exists()) {
            Log.trace("dist.rmi",1,
                      "Launching rmiregistry with "+"CLASSPATH="+jar.getPath());
            rmiProcess = Runtime.getRuntime().exec(
               new String[] { "rmiregistry", "-J-classpath", "-J"+jar.getPath()});
         } else {
            Log.trace("dist.rmi",1,
                      "Launching rmiregistry...");
            rmiProcess = Runtime.getRuntime().exec("rmiregistry");
         }
      } catch (IOException e) {
         Log.error("Could not start rmiregistry: "+e);
         e.printStackTrace();
      }
   }

   static Process rmiProcess = null;

   static {
      Runtime.getRuntime().addShutdownHook(
         new Thread() {
               public void run() {
                  if (rmiProcess!=null) {
                     System.out.println("Stopping rmiregistry...");
                     rmiProcess.destroy();
                  }
               }
            }
      );
   }

   /**
    * This method creates a new container and returns it.
    *
    * @param name  the container name
    * @return      the container reference
    */
   
   protected RemoteContainer newContainer(String name) {

      RMIRemoteContainer remCont = null;
      
      try { 
         remCont = new RMIRemoteContainer(verbose); 
      } catch(Exception e) { 
         e.printStackTrace(); 
         return null; 
      }
      
      try {
         registerContainer(remCont, name);
      } catch(Exception e) {
         throw new RuntimeException(e.toString());
      }
      remCont.getDelegate().setName(getFullHostName(name));

      return remCont.getDelegate();
   }
  

   /**
    * This method creates a new container, instantiates a given class, and
    * returns the container.
    *
    * @param name       the container name
    * @param className  the name of the class to instantiate
    * @return           the container reference
    */
   
   protected RemoteContainer newContainer(String name, String className) 
      throws RuntimeException {
   
      RMIRemoteContainer remCont = null;
      
      try { 
         remCont = new RMIRemoteContainer(className,verbose); 
      } catch( Exception e ) { 
         e.printStackTrace(); 
         return null; 
      }

      try {
         registerContainer(remCont, name);
      } catch(Exception e) {
         throw new RuntimeException(e.toString());
      }

      remCont.getDelegate().setName(getFullHostName(name));

      return remCont.getDelegate();
   }
   

   /**
    * This method register a container in the RMI registry.
    *
    * @param container  the container reference
    * @param name       the container name
    */
   
   protected void registerContainer(
   			RMIRemoteContainer container,
			String name
		  ) throws java.rmi.RemoteException {

         /** Register the container in the RMI registry */
         String fullName = getFullHostName(name);
	 
         try {
            Naming.rebind(fullName, container);
         } catch( java.net.MalformedURLException e ) {
            e.printStackTrace();
         }

         System.out.println(
            "--- JAC Distd (RMI): new container " + fullName + " ---"
         );
         System.out.println(
            "--- Default class repository: " + referenceContainerName + " ---"
         );
	 
   }

   /**
    * This method enters the event loop of
    * the underlying communication protocol.
    */

   public void run() {
      System.out.println( "--- JAC Distd (RMI) is running ---" );
   }

   /**
    * The is the main constructor.
    *
    * @param args  command line arguments
    */
   
   public RMIDistd(String[] args) { 
      super(args); 
   }
   
   public static void main(String[] args) { 
      new RMIDistd(args);
   }
   
}
