/*
  Copyright (C) 2002-2003 Renaud Pawlak <renaud@aopsys.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package org.objectweb.jac.ide;

import org.objectweb.jac.util.Log;
import org.objectweb.jac.util.Strings;
import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

public class Project extends ModelElement {

   public void generateCode() throws Exception {
      Log.trace("project","generate code in "+generationPath);
      if (generationPath!=null && !generationPath.equals("")) {
         Log.trace("project","creating project "+name+" code");
         CodeGeneration.createProjectCode(this,generationPath.getPath());
      } else { 
         Log.trace("project","cannot create "+name+" project code");
         throw new CannotGenerateException(
            "Project \""+name+"\" does not have a generation path");
      }
   }

   public void checkGenerationPath() throws CannotGenerateException {
      if (generationPath==null || generationPath.equals("")) {
         throw new CannotGenerateException(
            "Project \""+name+"\" does not have a generation path");
      }
   }

   public Errors compile() throws CannotCompileException, IOException, InterruptedException {
      Log.trace("project","compile code in "+generationPath);
      if (generationPath!=null && !generationPath.equals("")) {
         return CodeGeneration.compileProjectCode(this);
      } else {
         throw new CannotCompileException(
            "Project \""+name+"\" does not have a generation path");
      }
   }

   File compilerCommand = new File("javac");
   
   /**
    * Get the value of compilerCommand.
    * @return value of compilerCommand.
    */
   public File getCompilerCommand() {
      return compilerCommand;
   }
   
   /**
    * Set the value of compilerCommand.
    * @param v  Value to assign to compilerCommand.
    */
   public void setCompilerCommand(File  v) {
      this.compilerCommand = v;
   }

   String compilerOptions;
   public String getCompilerOptions() {
      return compilerOptions;
   }
   public void setCompilerOptions(String options) {
      this.compilerOptions = options;
   }
   
   File generationPath;
   
   /**
    * Get the value of generationPath.
    * @return value of generationPath.
    */
   public File getGenerationPath() {
      return generationPath;
   }
   
   /**
    * Set the value of generationPath.
    * @param v  Value to assign to generationPath.
    */
   public void setGenerationPath(File  v) {
      this.generationPath = v;
   }   

   Vector packages=new Vector();
   
   /**
    * Get the value of packages.
    * @return value of packages.
    */
   public List getPackages() {
      return packages;
   }
   
   public void addPackage(Package p) {
      packages.add(p);
   }
   
   public void removePackage(Package p) {
      packages.remove(p);
   }
   
   /**
    * Gets a package with a given name
    * @param packageName the requested package name
    */
   public Package getPackageByName(String packageName) {
      Log.trace("ide","getPackageByName "+packageName);
      Iterator it = packages.iterator();
      while (it.hasNext()) {
         Package pkg = (Package)it.next();
         if (pkg.getName().equals(packageName))
            return pkg;
      }
      return null;
   }


   /**
    * Find a package by its full name (my.package.ClassName)
    * @param pkgName the package name to find
    */
   public Package findPackage(String pkgName) {
      int dot = pkgName.indexOf('.');
      if (dot!=-1) {
         Package pkg = getPackageByName(pkgName.substring(0,dot));
         if (pkg!=null)
            return pkg.findPackage(pkgName.substring(dot+1));
         else 
            return null;
      } else {
         return getPackageByName(pkgName);
      }
   }

   Vector applications = new Vector();
   
   /**
    * Get the value of applications.
    * @return value of applications.
    */
   public List getApplications() {
      return applications;
   }
   
   public void addApplication(Application a) {
      applications.add(a);
      a.addAspectConfiguration(new AspectConfiguration("rtti"));
      a.setProject(this);
   }

   public void removeApplication(Application a) {
      applications.remove(a);
   }

   /** List of File */
   List classpath = new Vector();
   public List getClasspath() {
      return classpath;
   }
   public String getClasspathString() {
      return Strings.join(classpath,System.getProperty("path.separator"));
   }
   public void addClasspath(File path) {
      classpath.add(path);
   }
   public void removeClasspath(File path) {
      classpath.remove(path);
   }

   /**
    * Find a class by its full name (my.package.ClassName)
    * @param className the class name to find
    */
   public Class findClass(String className) {
      int dot = className.indexOf('.');
      if (dot!=-1) {
         String packageName = className.substring(0,dot);
         Package pkg = getPackageByName(packageName);
         if (pkg!=null)
            return pkg.findClass(className.substring(dot+1));
         else 
            return null;
      } else {
         return null;
      }
   }

   /**
    * Returns all classes of all packages of the project.
    */
   public Collection getClasses() {
      Vector classes = new Vector();
      Iterator it = packages.iterator();
      while (it.hasNext()) {
         Package pkg = (Package)it.next();
         classes.addAll(pkg.getAllClasses());
      }
      return classes;
   }
}
