/*
  Copyright (C) 2003 Laurent Martelli <laurent@aopsys.com>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package org.objectweb.jac.ide.swing;



import org.objectweb.jac.aspects.gui.swing.DefaultCompletionEngine;
import org.objectweb.jac.core.AspectComponent;
import org.objectweb.jac.core.parsers.acc.NonTerminal;
import org.objectweb.jac.core.parsers.acc.SyntaxElement;
import org.objectweb.jac.core.parsers.acc.Terminal;
import org.objectweb.jac.core.parsers.acc.ToolParserWrapper;
import org.objectweb.jac.core.rtti.AbstractMethodItem;
import org.objectweb.jac.core.rtti.ClassItem;
import org.objectweb.jac.core.rtti.FieldItem;
import org.objectweb.jac.core.rtti.MemberItem;
import org.objectweb.jac.ide.Class;
import org.objectweb.jac.ide.Field;
import org.objectweb.jac.ide.Package;
import org.objectweb.jac.ide.Project;
import org.objectweb.jac.ide.RelationRole;
import org.objectweb.jac.util.Log;
import java.io.StringReader;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import org.objectweb.jac.ide.Method;

/**
 * This class implements a completion engine for method acc configuration code of the UMLAF IDE. 
 */
public class AccCompletionEngine extends DefaultCompletionEngine {
   Project project;
   /**
    * Creates a new AccCompletionEngine using a given parser.
    * @param parser the parser to use
    */
   public AccCompletionEngine(ToolParserWrapper parser, Project project) {
      this.project = project;
      this.parser = parser;
   }

   SyntaxElement currentSyntaxElement;
   public SyntaxElement getCurrentSyntaxElement() {
      return currentSyntaxElement;
   }

   AspectComponent aspectInstance;
   /**
    * Sets the aspect instance associated with the completion engine
    * @param instance the AspectComponent instance
    */
   public void setAspectInstance(AspectComponent instance) {
      this.aspectInstance = instance;
   }

   ToolParserWrapper parser;
   public List getContextualChoices(String text, int position, 
                                    String writtenText) {
      NonTerminal elements = parser.parse(new StringReader(text),"");
      if (writtenText.length()>0)
         position = position+writtenText.length()-1;
      Log.trace("completion",2,"Syntax elements = "+elements);
      Log.trace("completion","position = "+position);
      Log.trace("completion","writtenText = "+writtenText);

      SyntaxElement se = parser.getSyntaxElementAt(position);
      currentSyntaxElement = se;
      Log.trace("completion","Syntax element = "+se);
      if (se!=null && (se.getName().equals("CONF_METHOD") || 
                       se.getName().equals("EOL") ||
                       se.getName().equals("class_block"))) {
         Log.trace("completion","Completing!");
         NonTerminal block = (NonTerminal)se.findParent("class_block");
         if (block!=null) {
            Terminal keyword = (Terminal)block.getChild("BLOCK_KEYWORD");
            if (keyword!=null) {
               // Completion inside a block
               if (keyword.getValue().equals("attribute")) {
                  Log.trace("completion","Completing in attribute block");
                  return aspectInstance.getConfigurationMethodsName(
                     FieldItem.class);
               } else if (keyword.getValue().equals("method")) {
                  Log.trace("completion","Completing in method block");
                  return aspectInstance.getConfigurationMethodsName(
                     AbstractMethodItem.class);
               } else if (keyword.getValue().equals("member")) {
                  Log.trace("completion","Completing in member block");
                  return aspectInstance.getConfigurationMethodsName(
                     MemberItem.class);
               } else if (keyword.getValue().equals("class")) {
                  Log.trace("completion","Completing in class block");
                  Vector result = new Vector();
                  result.addAll(
                     aspectInstance.getConfigurationMethodsName(ClassItem.class));
                  result.addAll(
                     aspectInstance.getConfigurationMethodsName(MemberItem.class));
                  result.addAll(Arrays.asList(
                     new String[] {"attribute","method","member"}));
                  return result;
               } else {
                  Log.trace("completion","Unknown block keyword");
                  return baseWords;
               }
            } else {
               Log.trace("completion","No keyword in block");
               return baseWords;
            }
         } else {
            Log.trace("completion","No block keyword");
            return baseWords;
         }
      } else {
         Terminal term = parser.getTerminalAt(position);
         Log.trace("completion","SyntaxElement before: "+term);
         if (term!=null && term.getName().equals("BLOCK_PARAM")) {
            return completeBlockParam(term);
         } else if (se==null) {
            Log.trace("completion","No current syntax element");
            return baseWords;
         }
      }
      return new Vector();
   }

   /**
    * Gets completion for a block parameter
    * @param term the Terminal syntax element to complete
    */
   protected List completeBlockParam(Terminal term) {
      Log.trace("completion","complete block param "+term);
      Vector result = new Vector();
      NonTerminal block = (NonTerminal)term.findParent("class_block");
      if (block!=null) {
         Terminal keyword = (Terminal)block.getChild("BLOCK_KEYWORD");
         if (keyword!=null) {
            if (keyword.getValue().equals("attribute")) {
               String className = getClassName(block);
               if (className!=null) {
                  Log.trace("completion",
                            "completing attribute for class "+className);
                  Class cl = project.findClass(className);
                  if (cl!=null) {
                     completeAttributeName(term.getValue(),cl,result);
                  }
               }
            } else if (keyword.getValue().equals("method")) {
               String className = getClassName(block);
               if (className!=null) {
                  Log.trace("completion",
                            "completing method for class "+className);
                  Class cl = project.findClass(className);
                  if (cl!=null) {
                     completeMethodName(term.getValue(),cl,result);
                  }
               }
            } else if (keyword.getValue().equals("member")) {
               String className = getClassName(block);
               if (className!=null) {
                  Log.trace("completion",
                            "completing member for class "+className);
                  Class cl = project.findClass(className);
                  if (cl!=null) {
                     completeAttributeName(term.getValue(),cl,result);
                     completeMethodName(term.getValue(),cl,result);
                  } else {
                     Log.warning("No such class in project: "+className);
                  }
               }
            } else if (keyword.getValue().equals("class")) {
               completeClassName(term.getValue(),result);
            } else {
               Log.trace("completion","BLOCK_KEYWORD is "+keyword.getValue());
            }
         } else {
            Log.trace("completion","No child BLOCK_KEYWORD "+term);
         }
      } else {
         Log.trace("completion","No parent class_block "+term);
      }

      return result;
   } 

   /**
    * Gets the className for a member,method or attribute block
    * @param block Non terminal of the block keyword
    * @return the class name, or null if it cannot be computed
    */
   protected String getClassName(NonTerminal block) {
      String className = null;
      block = (NonTerminal)block.getParent().findParent("class_block");
      if (block!=null) {
         Terminal keyword = (Terminal)block.getChild("BLOCK_KEYWORD");
         if (keyword!=null && keyword.getValue().equals("class")) {
            NonTerminal blockParams = 
               (NonTerminal)block.getChild("block_params");
            if (blockParams!=null) {
               Terminal param = (Terminal)blockParams.getChild(0);
               className = param.getValue();
            }
         }
      }
      return className;
   }

   protected void completeAttributeName(String start, Class cl, List result) {
      Iterator it = cl.getFields().iterator();
      while(it.hasNext()) {
         Field field = (Field)it.next();
         if (field.getName().startsWith(start)) {
            result.add(field.getGenerationName());
         }
      }
      it = cl.getNavigableRoles().iterator();
      while(it.hasNext()) {
         RelationRole role = (RelationRole)it.next();
         if (role.getGenerationName().startsWith(start)) {
            result.add(role.getGenerationName());
         }
      }
   }

   protected void completeMethodName(String start, Class cl, List result) {
      Iterator it = cl.getMethods().iterator();
      while(it.hasNext()) {
         Method method = (Method)it.next();
         if (method.getName().startsWith(start)) {
            result.add(method.getGenerationName());
         }
      }
   }

   protected void completeClassName(String start, List result) {
      Log.trace("completion","completing class name "+start);
      int dot = start.lastIndexOf('.');
      if (dot==-1) {
         Iterator it = project.getPackages().iterator();
         while (it.hasNext()) {
            Package pkg = (Package)it.next();
            if (pkg.getGenerationName().startsWith(start)) {
               result.add(pkg.getPPath());
            }
         }
         
      } else {
         String end = start.substring(dot+1);
         Package pkg = project.findPackage(start.substring(0,dot));

         if (pkg!=null) {
            Log.trace("completion","Looking for "+end+" in package "+pkg.getPPath());
            Iterator it = pkg.getClasses().iterator();
            while (it.hasNext()) {
               Class cl = (Class)it.next();
               if (cl.getGenerationName().startsWith(end)) {
                  Log.trace("completion","Found class "+cl.getFullName());
                  result.add(cl.getGenerationFullName());
               }
            }
            it = pkg.getSubPackages().iterator();
            while (it.hasNext()) {
               Package subPkg = (Package)it.next();
               if (subPkg.getGenerationName().startsWith(end)) {
                  Log.trace("completion","Found package "+subPkg.getPPath());
                  result.add(subPkg.getPPath());
               }
            }
         }
      }
   }
}

